!!$
!!$=head1 NAME
!!$
!!$ bwi-wet3d-cgstep.x - 2/3D one-way wave-equation migration velocity analysis
!!$   Written by Yaxun Tang
!!$   tang@sep.stanford.edu
!!$   Version 1.0
!!$   June, 2011
!!$
!!$=head2 SYNOPSIS FOR WAVEFIELD TOMOGRAPHY
!!$
!!$   bwi-wet3d-cgstep.x mode=tominv parameters
!!$
!!$=head3 PARAMETERS
!!$
!!$ csou=CSOU.H(input), source function
!!$      sep77 hypercube where n1=gx n2=gy n3=frequency n4=sx n5=sy
!!$      for point sources, set n1=1 o1=0 d1=1 d2=1 o2=0 d2=1 n4=1 o4=0 d4=1 n5=1 o5=0 d5=1
!!$      for 2-D, set n2=1 o2=0 d2=1 n5=1 o5=0 d5=1
!!$
!!$ crec=CREC.H(input), data to be migrated
!!$      sep77 hypercube where n1=gx n2=gy n3=frequency n4=sx n5=sy
!!$      for 2-D, set n2=1 o2=0 d2=1 n5=1 o5=0 d5=1
!!$
!!$ bvel=BVEL.H(input/output), input initial velocity model, overwritten by the updated model on output
!!$      sep77 hypercube where n1=x n2=y n3=z
!!$      for 2-D, set n2=1 o2=0 d2=1
!!$      the velocity model should be big enough to accommodate all sources and receivers
!!$
!!$ imsk=IMSK.H(input, optional), image mask, useful when imgmsk=y
!!$      sep77 hypercube where n1=x n2=y n3=z
!!$      for 2-D, set n2=1 o2=0 d2=1
!!$      the dimensions and samplings of x, y and z should be the same as those of BIMG.H 
!!$
!!$ gmsk=GMSK.H(input, optional), gradient mask, useful when grdmsk=y
!!$       sep77 hypercube where n1=x n2=y n3=z
!!$       for 2-D, set n2=1 o2=0 d2=1
!!$       the dimensions and samplings of x, y and z should be the same as those of BVEL.H
!!$
!!$ bimg=BIMG.H(output), a temporary image file, constantly overwritten by background image and perturbed image at each iteration
!!$      sep77 hypercube where n1=subsurface offset x n2=subsurface offset y n3=x n4=y n5=z
!!$
!!$ grdo=GRDO.H(output), gradient at current iteration, constantly overwritten at each iteration
!!$      sep77 hypercube where n1=x n2=y n3=z
!!$
!!$ grdp=GRDP.H(output), gradient at next iteration, constantly overwritten at each iteration
!!$      sep77 hypercube where n1=x n2=y n3=z
!!$
!!$ srch=SRCH.H(output), search direction at current iteration, constantly overwritten at each iteration
!!$      sep77 hypercube where n1=x n2=y n3=z
!!$
!!$ mmov=MMOV.H(output, optional), inverted velocity model as a function of iteration, useful when modmov=y
!!$      sep77 hypercube where n1=x n2=y n3=z n4=iteration
!!$
!!$ fobj=FOBJ.H(output, optional), objective function value as a function of iteration, useful when objmov=y
!!$      sep77 hypercube where n1=iteration
!!$
!!$ dipx=DIPX.H(output, optional), dip field of the background image in x direction, constantly overwritten at each iteration, useful when wntdip=y
!!$      sep77 hypercube where n1=x n2=y n3=z
!!$
!!$ dipy=DIPY.H(output, optional), dip field of the background image in y direction, constantly overwritten at each iteration, useful when wntdip=y
!!$      sep77 hypercube where n1=x n2=y n3=z
!!$
!!$ bwfs=BWFS.H(output, optional), source wavefield, useful when wntwfs=y
!!$      sep77 hypercube where n1=x n2=y n3=frequency n4=z
!!$
!!$ bwfr=BWFR.H(output, optional), receiver wavefield, useful when wntwfr=y
!!$      sep77 hypercube where n1=x n2=y n3=frequency n4=z
!!$
!!$ radian=[y]/n, whether the input frequency is in radian or not
!!$
!!$ offset=y/[n], y for marine geometry where the receivers are in offset coordinates
!!$
!!$ comsou=y/[n], y for composite or areal source function
!!$               n for point source function
!!$
!!$ smtsou=y/[n], whether or not to spatially smooth the point source function, useful when comsou=n
!!$
!!$ stkimg=[y]/n, if the DIMG.H input for back-projection hasn't been stacked over source axis, set stkimg=y, otherwise n
!!$
!!$ bsctyp=[fourie]/tayexp, type of the Born scattering operator
!!$                         fourie - Fourier domain PSPI-type of implementation,
!!$                                  in this case, use nvrf to set the # of reference velocities
!!$                                  used for calculating the scattering operator
!!$                         tayexp - Born scattering operator based on Taylor series expansion
!!$
!!$ bscnop=[2], number of terms used in Taylor series expansion, useful when bsctyp=tayexp
!!$
!!$ bsccoe=[optimz]/taylor, coefficients for Taylor series expansion, useful when bsctyp=tayexp
!!$                         optimz - optimized coefficients
!!$                         taylor - coefficients based on Taylor expansion
!!$
!!$ srcbnd=y/[n], boundary condition for source function, which changes dipole sources into monopole
!!$
!!$ vsfc=[1500], surface velocity for source boundary condition, useful when srcbnd=y
!!$
!!$ nxpad_beg=[0], zero padding (# of samples) at the beginning of x
!!$
!!$ nxpad_end=[0], zero padding (# of samples) at the end of x
!!$
!!$ nypad_beg=[0], zero padding (# of samples) at the beginning of y
!!$
!!$ nypad_end=[0], zero padding (# of samples) at the end of y
!!$
!!$ ntpx=[1], taper length (# of samples) in x
!!$
!!$ ntpy=[1], taper length (# of samples) in y
!!$
!!$ oper=phsft/ssfou/impfd/foufd/phspi/[ssfpi], extrapolator type
!!$                                             phsft: phase shift
!!$                                             ssfou: split-step Fourier
!!$                                             impfd: implicit finite difference
!!$                                             foufd: Fourier finite difference
!!$                                             phspi: phase-shift plus interpolation
!!$                                             ssfpi: split-step Fourier plus interpolation
!!$
!!$ nvrf=[1], number of reference velocities, useful when oper=phspi or ssfpi
!!$
!!$ pdip=[45], propagation dip for finite difference, useful when oper=impfd or foufd
!!$            for oper=impfd, pdip can be chosen from 15/45/65/79/80/87/89/90,
!!$                            wide angle propagation is achieved by Zaitian Ma's 
!!$                            higher-order splitting method
!!$            for oper=foufd, pdip can be chosen from 45/65/75 
!!$
!!$ trick=[0.14], 1/6 trick for finite difference, useful when oper=impfd or foufd
!!$
!!$ bndpr=[0], boundary condition (real part) for tridiagonal system, useful when oper=impfd or foufd
!!$
!!$ bndpi=[0], boundary condition (imaginary part) for tridiagonal system, useful when oper=impfd or foufd
!!$
!!$ bullet=y/[n], use bullet-proof finite-difference shceme or not, useful when oper=impfd or foufd
!!$
!!$ errflt=[y]/n, apply Li's compensation filter or not, useful when oper=impfd or foufd
!!$
!!$ wntwfs=y/[n], whether or not to output the source wavefield
!!$
!!$ wntwfr=y/[n], whether or not to output the receiver wavefield
!!$
!!$ nws=[1], number of frequencies to be held in memory
!!$          set nws >= OMP_NUM_THREAD for better performance
!!$
!!$ report=[y]/n, whether or not to report progress
!!$
!!$ memchk=[y]/n, memory check or not, set memchk=n to actually run the program
!!$
!!$ imgfwd_type=[bckward]/adjoint, extrapolation type for modeling/migration
!!$
!!$ tombwd_type=bckward/[adjoint], extrapolation type for tomography
!!$
!!$ nhx=[1], number of subsurface offset in x
!!$
!!$ ohx=[0], origin of subsurface offset in x
!!$
!!$ dhx=[dgx], sampling rate of subsurface offset in x
!!$
!!$ nhy=[1], number of subsurface offset in y
!!$
!!$ ohy=[0], origin of subsurface offset in y
!!$
!!$ dhy=[dgy], sampling rate of subsurface offset in y
!!$
!!$ image_xmin image_xmax, minimum and maximum x values for image output
!!$
!!$ image_ymin image_ymax, minimum and maximum y values for image output
!!$
!!$ image_zmin image_zmax, minimum and maximum z values for image output
!!$
!!$ niter=[1], number of nonlinear iterations
!!$
!!$ nsrc=[3], number of step-length trials for line search at each iteration
!!$
!!$ alph=[20], initial step length
!!$
!!$ hcut=[0], the truncation point for offset-domain DSO, useful when objtyp=offdso or combos
!!$ 
!!$ dsotyp=[1]/2, DSO type, useful when objtyp=offdso or combos
!!$               1 for |h|
!!$               2 for a Gaussian taper function
!!$
!!$ dsorms=y/[n], whether or not to noramlize DSO with an image RMS amplitudes, useful when objtyp=offdso or combos
!!$
!!$ objtyp=[imgpow]/offdso/combos, type of objective function
!!$                                imgpow - image stack power maximization, in this case, set nhx=1 ohx=0 dhx=1 nhy=1 ohy=0 dhy=1
!!$                                offdso - subsurface-offset-domain DSO
!!$                                combos - combination of image stack power maximization and DSO
!!$
!!$ objeps=[1.], trade-off parameter to adjust the relative importance of DSO and image stack power (ISP)
!!$              Fobj = Fobj_DSO + objeps*Fobj_ISP, useful when objtyp=combos
!!$
!!$ method=[cgstep]/sdstep, optimization algorithm
!!$                         cgstep - nonlinear conjugate gradient (Polak-Ribiere scheme)
!!$                         sdstep - steepest descent
!!$
!!$ smtgrd=y/[n], whether or not to precondition (smooth) the gradient
!!$
!!$ smttyp=dipflt/[triflt]/spline, gradient preconditioning (smoothing) method
!!$                                dipflt - dip filter
!!$                                triflt - triangle
!!$                                spline - B-spline
!!$
!!$ scheme=[UW2x3]/LW2x3, finite-difference schemes for dip filters, usefule when smttyp=dipflt
!!$                       UW2x3 - upwind scheme
!!$                       LW2x3 - Lax-Wendroff scheme
!!$
!!$ rectx_dip=[1], triangle filter length in x for dip smoothing, useful when smttyp=dipflt
!!$
!!$ recty_dip=[1], triangle filter length in y for dip smoothing, useful when smttyp=dipflt 
!!$
!!$ rectz_dip=[1], triangle filter length in z for dip smoothing, useful when smttyp=dipflt
!!$
!!$ dip_min=[0.], minimum dip value (degree) for tabulating dip filters, useful when smttyp=dipflt
!!$
!!$ dip_max=[0.], maximum dip value (degree) for tabulating dip filters, useful when smttyp=dipflt
!!$
!!$ ndip=[1], number of dip filters to be precalculated, useful when smttyp=dipflt
!!$
!!$ dipitr=[100], number of iterations for Wilson-Burg spectral factorization for constructing dip filters, useful when smttyp=dipflt
!!$
!!$ dipesp=[0.001], damping factors for spectral factorization, useful when smttyp=dipflt
!!$
!!$ na=[5], number of coefficients (3*na) for dip filters on a helix, useful when smttyp=dipflt
!!$
!!$ wntdip=y/[n], whether or not to output the dip fields, useful when smttyp=dipflt
!!$
!!$ rectx_grd=[1], triangle filter length in x for gradient smoothing, useful when smttyp=triflt
!!$
!!$ recty_grd=[1], triangle filter length in y for gradient smoothing, useful when smttyp=triflt
!!$
!!$ rectz_grd=[1], triangle filter length in z for gradient smoothing, useful when smttyp=triflt
!!$
!!$ xk_min=[0] xk_max=[0], minimum and maximum x values for B-splne nodes, useful when smttyp=spline
!!$
!!$ yk_min=[0] yk_max=[0], minimum and maximum y values for B-splne nodes, useful when smttyp=spline
!!$
!!$ zk_min=[0] zk_max=[0], minimum and maximum z values for B-splne nodes, useful when smttyp=spline
!!$
!!$ movmov=y/[n], whether or not to output velocity model movie
!!$
!!$ objmov=y/[n], whether or not to output objective function movie
!!$
!!$=head1 COMMENTS
!!$
!!$ Report bugs to Yaxun Tang, tang@sep.stanford.edu
!!$
!!$=cut

program main
  use bwi_kernel_mod
  use bwi_restom_mod
  use bwi_grdprc_mod
  use bwi_search_mod
  use bwi_diphlx_mod
  implicit none

  type(dat_grid)                                  :: dgrid
  type(pro_grid)                                  :: pgrid
  type(vel_grid)                                  :: vgrid
  type(img_grid)                                  :: igrid
  type(hes_grid)                                  :: hgrid
  type(bwi_ftag)                                  :: ftag_ 
  type(axis)                                      :: aiter, adip
  type(axis)                                      :: axk, ayk, azk
  type(helix)    , dimension(:)     , allocatable :: dfcx , dfcy 
  real           , dimension(:)     , allocatable :: dfcx0, dfcy0
  integer        , dimension(:,:,:) , allocatable :: selx, sely
  integer                                         :: stat, iter, niter, nsrc, dsotyp, dipitr, siter, na, idip
  integer                                         :: rectx_grd, recty_grd, rectz_grd, nsmt_grd
  integer                                         :: rectx_dip, recty_dip, rectz_dip, nsmt_dip
  real                                            :: xk_min, yk_min, zk_min
  real                                            :: xk_max, yk_max, zk_max
  real                                            :: alph, hcut, objeps, dipeps
  real                                            :: dip_min, dip_max  
  double precision, dimension(:)    , allocatable :: fobj
  real            , dimension(:)    , allocatable :: vmod
  real            , dimension(:,:,:), allocatable :: dipx, dipy
  character(len=6)                                :: objtyp, method, smttyp
  character(len=5)                                :: scheme 
  logical                                         :: wntdip, smtgrd, modmov, objmov, switch, dsorms, grdfix
  logical, parameter                              :: T = .true., F = .false.
  character(len=256)                              :: fobj_, mmov_, dipx_, dipy_

  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  fobj_ = "fobj"
  mmov_ = "mmov"
  dipx_ = "dipx"
  dipy_ = "dipy"

  call from_param("niter" , niter , 1       )
  call from_param("nsrc"  , nsrc  , 3       )
  call from_param("alph"  , alph  , 20.     )
  call from_param("hcut"  , hcut  , 20.     )
  call from_param("dsotyp", dsotyp, 1       )
  call from_param("dsorms", dsorms, .false. )
  call from_param("objtyp", objtyp, "imgpow") 
  call from_param("switch", switch, .false. )
  call from_param("method", method, "cgstep")
  call from_param("wntdip", wntdip, .false. ) 
  call from_param("smtgrd", smtgrd, .false. )
  call from_param("smttyp", smttyp, "triflt")
  call from_param("scheme", scheme, "UW2x3" )
  call from_param("objeps", objeps, 1.      ) 
  call from_param("modmov", modmov, .false. )
  call from_param("objmov", objmov, .false. )
  call from_param("grdfix", grdfix, .false. )

  if (switch) then
     call from_param("siter", siter, 1)
     grdfix = .true.
  end if

! initialize Born wavefield tomography kernel
! -------------------------------------------
  write(0,*) "initialzing the kernel"
  stat = bwi_kernel_init(dgrid, pgrid, vgrid, igrid, hgrid, ftag_)

  if (pgrid%mode/="tominv") call seperr("ERROR: WRONG MODE")

  if (smtgrd) then
     select case (smttyp)
       case ("dipflt")
            call from_param("rectx_dip", rectx_dip, 1    )
            call from_param("recty_dip", recty_dip, 1    )
            call from_param("rectz_dip", rectz_dip, 1    )
            call from_param("nsmt_dip" , nsmt_dip , 1    )
            call from_param("dip_min"  , dip_min  , 0.   )
            call from_param("dip_max"  , dip_max  , 0.   )
            call from_param("ndip"     , adip%n   , 1    )
            call from_param("dipitr"   , dipitr   , 100  )
            call from_param("dipeps"   , dipeps   , 0.001)
            call from_param("na"       , na       , 5    )
            if (adip%n==1) then
               adip%o = dip_min
               adip%d = 1.
            else
               adip%o = dip_min
               adip%d = (dip_max-dip_min)/(adip%n-1)
            end if
            adip%n = adip%n + 1 
!           tabulate dip filters in x
!           -------------------------
            if (vgrid%ax%n>1) then
               allocate(dfcx(adip%n), dfcx0(adip%n))
               allocate(dipx(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               allocate(selx(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               do idip=1, adip%n
                  call allocatehelix(dfcx(idip), 3*na)
               end do
               stat = bwi_diphlx_init(adip, dipeps, vgrid%ax%d, vgrid%az%d, vgrid%ax%n, na, dipitr, scheme)
               stat = bwi_diphlx_oper(dfcx, dfcx0)
               stat = bwi_diphlx_clos()
            end if

!           tabulate dip filters in y
!           -------------------------
            if (vgrid%ay%n>1) then
               allocate(dfcy(adip%n), dfcy0(adip%n))
               allocate(dipy(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               allocate(sely(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               do idip=1, adip%n
                  call allocatehelix(dfcy(idip), 3*na)
               end do
               stat = bwi_diphlx_init(adip, dipeps, vgrid%ay%d, vgrid%az%d, vgrid%ay%n, na, dipitr, scheme)
               stat = bwi_diphlx_oper(dfcy, dfcy0)
               stat = bwi_diphlx_clos()
            end if           

            if (wntdip) then
               if (vgrid%ax%n>1) then
                  call auxinout(dipx_)
                  call houaxis(dipx_, vgrid%ax, 1)
                  call houaxis(dipx_, vgrid%ay, 2)
                  call houaxis(dipx_, vgrid%az, 3)
               end if
               if (vgrid%ay%n>1) then
                  call auxinout(dipy_)
                  call houaxis(dipy_, vgrid%ax, 1)
                  call houaxis(dipy_, vgrid%ay, 2)
                  call houaxis(dipy_, vgrid%az, 3)
               end if
            end if

       case ("triflt")
            call from_param("rectx_grd", rectx_grd, 1)
            call from_param("recty_grd", recty_grd, 1)
            call from_param("rectz_grd", rectz_grd, 1)
            call from_param("nsmt_grd" , nsmt_grd , 1) 

       case ("spline")
            call from_param("xk_min",xk_min,0.); call from_param("xk_max",xk_max,0.); call from_param("nxk",axk%n,1)
            call from_param("yk_min",yk_min,0.); call from_param("yk_max",yk_max,0.); call from_param("nyk",ayk%n,1)
            call from_param("zk_min",zk_min,0.); call from_param("zk_max",zk_max,0.); call from_param("nzk",azk%n,1)
            xk_min = vgrid%ax%o + (int(1.5+(xk_min-vgrid%ax%o)/vgrid%ax%d)-1)*vgrid%ax%d
            xk_max = vgrid%ax%o + (int(1.5+(xk_max-vgrid%ax%o)/vgrid%ax%d)-1)*vgrid%ax%d
            yk_min = vgrid%ay%o + (int(1.5+(yk_min-vgrid%ay%o)/vgrid%ay%d)-1)*vgrid%ay%d
            yk_max = vgrid%ay%o + (int(1.5+(yk_max-vgrid%ay%o)/vgrid%ay%d)-1)*vgrid%ay%d
            zk_min = vgrid%az%o + (int(1.5+(zk_min-vgrid%az%o)/vgrid%az%d)-1)*vgrid%az%d
            zk_max = vgrid%az%o + (int(1.5+(zk_max-vgrid%az%o)/vgrid%az%d)-1)*vgrid%az%d
            axk%o = xk_min; axk%d = 1.
            ayk%o = yk_min; ayk%d = 1.
            azk%o = zk_min; azk%d = 1.
            if (axk%n>1) then
               axk%o = xk_min; axk%d = (xk_max-xk_min)/(axk%n-1)
            end if
            if (ayk%n>1) then
               ayk%o = yk_min; ayk%d = (yk_max-yk_min)/(ayk%n-1)
            end if
            if (azk%n>1) then
               azk%o = zk_min; azk%d = (zk_max-zk_min)/(azk%n-1)
            end if
       case default
            write(0,*)"ERROR: smttyp = [triflt]/dipflt/spline"
            STOP
     end select
  end if

  aiter%n = niter + 1
  aiter%d = 1
  aiter%o = 0.

  allocate(fobj(0:niter))
  fobj = dble(0.)

  if (modmov) then
     call auxinout(mmov_) 
     call houaxis(mmov_, vgrid%ax, 1)
     call houaxis(mmov_, vgrid%ay, 2)
     call houaxis(mmov_, vgrid%az, 3)
     call houaxis(mmov_, aiter   , 4)
     allocate(vmod(vgrid%ax%n*vgrid%ay%n*vgrid%az%n)) 
  end if

  if (objmov) then
     call auxinout(fobj_)
     call houaxis(fobj_, aiter, 1)
  end if

! migrate with initial model
! --------------------------
  write(0,*) "generating background image"
  stat = bwi_kerimg_oper(T, F, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_)

! compute the imaging hessian
! --------------------------- 
  if (hgrid%wnthes) then
     write(0,*) "computing imaging Hessian"
     stat = bwi_kerhes_oper(F, ftag_%sfnc_, ftag_%gsou_, ftag_%grec_, ftag_%bvel_, ftag_%hess_)

     write(0,*) "normalizing Hessian"
     stat = bwi_hesnor_oper(ftag_%hess_, igrid, hgrid)
  end if

! compute the dip field, if requested
! -----------------------------------
  if (smtgrd .and. smttyp=="dipflt") then
     write(0,*)"update dip field"
     stat = bwi_update_dips(ftag_%bimg_, dipx, dipy, selx, sely, igrid, vgrid, adip, rectx_dip, recty_dip, rectz_dip, nsmt_dip)
     if (wntdip) then
        if (vgrid%ax%n>1) then
           call sseek(dipx_, 0, 0)
           call sep_write(dipx, dipx_)
        end if
        if (vgrid%ay%n>1) then
           call sseek(dipy_, 0, 0)
           call sep_write(dipy, dipy_)
        end if 
     end if
  end if

! compute the residual and objective function with initial model
! --------------------------------------------------------------
  write(0,*) "compute the image residual and objective function value"
  stat = bwi_restom_oper(ftag_%bimg_, ftag_%hess_, ftag_%imsk_, igrid, hgrid, fobj(0), hcut, dsotyp, dsorms, objtyp, objeps, grdfix)

! compute the gradient with the initial model
! -------------------------------------------
  write(0,*) "compute the gradient"
  stat = bwi_kertom_oper(T, F, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%grdo_, ftag_%bimg_)

! smooth the gradient if requested
! --------------------------------
  if (smtgrd) then
     write(0,*) "smooth the gradient: ", smttyp
     select case (smttyp)
       case ("dipflt")
            stat = bwi_dipflt_oper(ftag_%grdo_, dfcx, dfcy, dfcx0, dfcy0, selx, sely, vgrid)
       case ("triflt")
            stat = bwi_triflt_oper(ftag_%grdo_, vgrid, rectx_grd, recty_grd, rectz_grd, nsmt_grd) 
       case ("spline")
            stat = bwi_spline_oper(ftag_%grdo_, vgrid, axk, ayk, azk)
     end select
  end if

! apply gradient mask
! -------------------
  if (vgrid%grdmsk) then
     write(0,*)"apply gradient mask"
     stat = bwi_grdmsk_oper(ftag_%grdo_, ftag_%gmsk_, vgrid)
  end if

! scaling and normalizing the gradient
! ------------------------------------
  write(0,*) "scale and normalize the data-fit gradient"
  stat = bwi_gscale_oper(ftag_%grdo_, vgrid)

! initialize the search direction
! -------------------------------
  write(0,*) "initialize the search direction"
  stat = bwi_sdstep_oper(ftag_%srch_, ftag_%grdo_, vgrid)

  if (modmov) then
     call sseek(ftag_%bvel_, 0, 0)
     call sep_read(vmod, ftag_%bvel_)
     call sep_write(vmod, mmov_)
  end if

  write(0,*) "-------------------"
  write(0,*) "  BEGIN ITERATION  "
  write(0,*) "-------------------"

  write(0,*) "iter = ", iter, fobj(0), alph, dsotyp

  do iter=1, niter

!    perform line search to update the model
!    ---------------------------------------
     stat = bwi_linsec_oper(ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_, ftag_%hess_, ftag_%srch_, ftag_%imsk_, &
                            vgrid, igrid, hgrid, nsrc, alph, fobj(iter-1), hcut, dsotyp, dsorms, objtyp, objeps, grdfix)

!    switch the DSO objective function
!    ---------------------------------
     if (switch .and. iter>siter) grdfix = .false.

!    migrate with current model
!    --------------------------
     stat = bwi_kerimg_oper(T, F, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_)

!    if Hessian is requested
!    -----------------------
     if (hgrid%wnthes) then
!       compute the imaging hessian
!       ---------------------------
        stat = bwi_kerhes_oper(F, ftag_%sfnc_, ftag_%gsou_, ftag_%grec_, ftag_%bvel_, ftag_%hess_)
!       apply the imaging Hessian
!       -------------------------
        stat = bwi_hesnor_oper(ftag_%hess_, igrid, hgrid)
     end if

!    calculate the dip field, if requested
!    -------------------------------------
     if (smtgrd .and. smttyp=="dipflt") then
        stat = bwi_update_dips(ftag_%bimg_, dipx, dipy, selx, sely, igrid, vgrid, adip, rectx_dip, recty_dip, rectz_dip, nsmt_dip)
        if (wntdip) then
           if (vgrid%ax%n>1) then  
              call sseek(dipx_, 0, 0)
              call sep_write(dipx, dipx_)
           end if
           if (vgrid%ay%n>1) then
              call sseek(dipy_, 0, 0)
              call sep_write(dipy, dipy_)
           end if
        end if
     end if 

!    compute the residual and objective function with current model
!    --------------------------------------------------------------
     stat = bwi_restom_oper(ftag_%bimg_, ftag_%hess_, ftag_%imsk_, igrid, hgrid, fobj(iter), hcut, dsotyp, dsorms, objtyp, objeps, grdfix)

!    compute the gradient with the current model
!    -------------------------------------------
     stat = bwi_kertom_oper(T, F, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%grdp_, ftag_%bimg_)

!    smooth the gradient
!    -------------------
     if (smtgrd) then
        select case (smttyp)
          case ("dipflt")
               stat = bwi_dipflt_oper(ftag_%grdp_, dfcx, dfcy, dfcx0, dfcy0, selx, sely, vgrid)
          case ("triflt")
               stat = bwi_triflt_oper(ftag_%grdp_, vgrid, rectx_grd, recty_grd, rectz_grd, nsmt_grd)
          case ("spline")
               stat = bwi_spline_oper(ftag_%grdp_, vgrid, axk, ayk, azk)
        end select
     end if

!    apply gradient mask
!    ------------------- 
     if (vgrid%grdmsk) then
        stat = bwi_grdmsk_oper(ftag_%grdp_, ftag_%gmsk_, vgrid)
     end if

!    scale and normalize the gradient
!    --------------------------------
     stat = bwi_gscale_oper(ftag_%grdp_, vgrid)

!    get the search direction
!    ------------------------
     if (method=="cgstep" .and. switch .and. iter==siter) then
        stat = bwi_cgstep_oper(ftag_%srch_, ftag_%grdo_, ftag_%grdp_, vgrid, .true.)

     else
        select case (method)
          case ("sdstep")
               stat = bwi_sdstep_oper(ftag_%srch_, ftag_%grdp_, vgrid)
          case ("cgstep")
               stat = bwi_cgstep_oper(ftag_%srch_, ftag_%grdo_, ftag_%grdp_, vgrid, .false.)
        end select
     end if

     if (modmov) then
        call sseek(ftag_%bvel_, 0, 0)
        call sep_read(vmod, ftag_%bvel_)
        call sep_write(vmod, mmov_)
     end if 

     write(0,*) "iter = ", iter, fobj(iter), alph, dsotyp

  end do

  if (objmov) call sep_write(real(fobj), fobj_)


  write(0,*)"***********************"
  write(0,*)"      Finished!!!      "
  write(0,*)"***********************"

  call sep_end_prog()

  call exit()

end program
