!!---------------------------------------------------------------------
!!.. Wave-equation-based inversion modules
!!.. OpenMP parallelized
!!.. Version 8.0
!!.. 1. This version includes the forward and adjoint operators
!!.. 2. This version also parallelizes the imaging condition
!!.. 3. This version uses the same sign for both source and receiver
!!..    wavefield and takes the conjugate of the source wavefield
!!..    after it is read from the disk immediately 
!!.. 4. This version includes complex image output/input (1/28/2009)
!!.. 5. This version supports output/input unstacked image 
!!..    (gather with surface parameters) (2/6/2009)
!!.. 6. This version fully passes dot-product, even with taper added
!!..    on (1/14/2010)
!!.. 7. This version supports output image on a coarser grid (1/14/2010)
!!.. 8. This version implements bulletproof implicit finite difference
!!..    and Fourier finite difference (8/2010)
!!.. 9. This version includes forward, adjoint and backward extrapolation
!!..    of wavefields, fully passes dot-product test for variable velocities 
!!..    with adjoint extrapolation (8/2010)
!!.. 10. This version implements phase error compensation filter for
!!..     for implicit finite difference (8/2010)
!!.. 11. This version preconditions the gradient with non-stationary
!!..     dip filters (10/2010)
!!---------------------------------------------------------------------
!!.. Author: Yaxun Tang, tang@sep.stanford.edu
!!---------------------------------------------------------------------
!!.. Note: 1. To make full usage of the threads, 
!!..          it is better to keep nws >= nth 
!!..       2. For threads safty purpose, the memory allocation is
!!..          considerably increased compared to serial implementation.
!!..          Mem of parallel ~ Mem of serial * nth
!!---------------------------------------------------------------------
!!  Yi's change (Nov, 2011) --bwi_kernel_oper
!! 1. wfdexp(intepolation not included, and only for split step)
!! 2. add bwi_qref_mod.f90 (uniformly distributed)
!! 3. bwi_fileio_mod (qgrid%index_provel_ymin?)
!! 4. bwi_report_mod
!!---------------------------------------------------------------------
!!---------------------------------------------------------------------
!! Yi' change

!!----------------Forward modeling -changed Jan 20, 2012--------------------------------------

!! 1.bwi_struct_mod 
!!	Add: bq_, dq_
!!	1.1 base_struct_mod
!!		Add:q_grid(type):proq,imgq,
!!	    	    pro_grid:nvrf,nqrf	

!! 2. bwi_report_mod
!!	Add:q_grid

!! 3. bwi_wfdexp_mod
!!	case:phsft, ssfou, phspi, ssfpi
!!	3.1 bwi_ssrepi_mod
!!		3.1.1 	iwrk2d -> ivwrk2d, iqwrk2d
!!		3.1.2   do ivrf, then do iqvrf
!!		3.1.3   ivwrk2d==ivrf .and. iqwrk2d==iqr
!!	3.2 bwi_extrpl_mod
!!		3.2.1 phase_shift
!!			vref -> cref
!!			wov=ws/cvref
!!		3.2.2 ssf_correct
!!			vref -> cvref
!!			vmod -> cvmod
!!		3.2.3 ssr_phs
!!			Add qref
!!		3.2.4 ssr_ssf
!!			Add qmod_slice, qref
!!		3.2.5 ssr_ffd 
!!			Add Q (calling and defination only to prevent error report)
!! 4. bwi_fileio_mod
!!	Add bwi_fio_q, bwi_fio_qmod_updt, bwi_fio_dq_init
!!	sreed_window(vmod)->sreed(vmod); sreed_window(vmod)->sreed(vmod) Jan 31, 2012

!! 5. bwi_histos_mod
!!	bwi_histos_imgfwd: subroutine, type-qgrid, hinaxis, character(len=*),
!!	bwi_histos_imgadj: subroutine, type-qgrid, hinaxis, character(len=*),

!! 6. comment:bwi_kertom_oper, bwi_tom1st, bwi_kerhes_oper, bwi_hes1st
!! 7. bwi_glocal_mod:  add  base_qgrd_gen3d and call base_qgrd_gen3d(qgrid, pgrid, igrid)

!!---------------- Migration -changed Mar 3,2012----------------
!!-----------------Q Compensation -Apr 5,2012-------------------
!! 1. kernal:    
!!	 bwi_kerimg_oper (qcomp, qcomp_clip): bwi_img1st(qcomp,qcomp_clip), define qcomp and qcomp_clip
!!	 bwi_img1st(qcomp,qcomp_clip): define qcomp and qcomp_clip, bwi_wfdexp_oper(qcomp, qcomp_clip)
!! 2. bwi_wfdexp_mod
!!	1) define qcomp and qcomp_clip
!!	2) add qcomp and qcomp_clip in ssr_phs, ssr_ssf, ssr_epi, ssr_epi
!! 3. bwi_extrpl_mod
!!	1) ssr_phs: define, phase_shift
!!	2) ssr_ssf: define, phase_shift
!!	3) phase_shift
!!	4) ssr_ffd: phase_shift(calling and defination only to prevent error report)
!!	5) Notice: ssf_correct needs change?
!! 4.bwi_ssrepi_mod
!! 	1)ssr_epi (qcomp, qcomp_clip): define, stat=FXOP
!!	2)Fxop(qcomp, qcomp_clip): define
!!------------------------------------------------

!!-----------------Q tmography -Apr 5,2012-------------------
!!! 1.bwi_histos_mod:
!!!		     bwi_histos_tomfwd:subroutine, type, character(len=*), call hinaxis, seperr, change ax->bx for dvel(?)
!!! 		     bwi_histos_tomadj:subroutine, type, character(len=*), call hinaixs, houtaxis
!!! 		     bwi_histos_tominv:subroutine, type, character(len=*), call hinaixs
!!! 2. kernal: 
!!! 		select case: tomfwd, tomadj, tominv
!!! 		bwi_memory_chck: tomfwd, tomadj, tominv
!!! 		bwi_arrays_init: tomfwd, tomadj, tominv
!!!  		uncomment:bwi_kertom_oper, bwi_tom1st
!!! 3. bwi-wet3d:
!!!		Add bq, dq, qcomp/qcomp_clip

!!-----------------Q tmography and inversion -Apr 5,2012-------------------
!! 1. remove all dvel

!!-----------------convert Q into log
!! 1. all Q is considered as logQ

!! ------------------Add p1, p2, p3, p4 --------
!! 1. kernel
!! 2. wdexp
!! 3. expl

!!!----------------Add Q to Hessian -Oct 16, 2013-----------
!! 1.  1. source intensity; 2. receiver intensity
module bwi_kernel_mod
  use  base_blkcal_mod
  use  base_spline_mod
  use  bwi_struct_mod
  use  bwi_histos_mod
  use  bwi_glocal_mod
  use  bwi_fileio_mod
  use  bwi_velref_mod
  use  bwi_wfdexp_mod
  use  bwi_imgcon_mod
  use  bwi_fftwv3_mod
  use  bwi_tapers_mod
  use  bwi_sbcond_mod
  use  bwi_bornsc_mod
  use  bwi_rdmgen_mod
  use  bwi_hescpt_mod
  use  bwi_report_mod
  use  bwi_sbcond_mod
  use  bwi_qref_mod		!!! Add qref
  implicit none
  type(dat_grid)                              , private :: dgrid    !!.. receiver grid
  type(vel_grid)                              , private :: vgrid    !!.. velocity grid
  type(q_grid)                                , private :: qgrid    !!.. q grid
  type(pro_grid)                              , private :: pgrid    !!.. propagation grid
  type(img_grid)                              , private :: igrid    !!.. image grid
  type(out_grid)                              , private :: ogrid    !!.. image output grid
  type(hes_grid)                              , private :: hgrid    !!.. Hessian grid

  type(block)                                 , private :: wblock

  type(bwi_ftag)                              , private :: ftag_

  real     , dimension(:,:)      , allocatable, private :: bvel     !!.. velocity model, backgorund
 ! real     , dimension(:,:,:)    , allocatable, private :: dvel     !!.. perturbed velocity model
  real     , dimension(:)        , allocatable, private :: bvrf     !!.. reference velocity
  real     , dimension(:,:)      , allocatable, private :: bq     !!.. q model, backgorund
  real     , dimension(:,:,:)    , allocatable, private :: dq     !!.. perturbed q model
  real     , dimension(:)        , allocatable, private :: bqrf     !!.. reference q
  real     , dimension(:,:,:,:,:), allocatable, private :: rimg     !!.. output image, real
  real     , dimension(:,:,:,:,:), allocatable, private :: hess
  real     , dimension(:,:)      , allocatable, private :: imsk

  complex  , dimension(:)        , allocatable, private :: sfnc
  complex  , dimension(:,:,:,:,:), allocatable, private :: cimg     !!.. output image, complex
  complex  , dimension(:,:)      , allocatable, private :: crec
  complex  , dimension(:,:)      , allocatable, private :: csou
  complex  , dimension(:,:,:,:)  , allocatable, private :: bwfs, bwfr
  complex  , dimension(:,:,:,:)  , allocatable, private :: dwfs, dwfr

  complex  , dimension(:,:,:)    , allocatable, private :: bwfs_slice, bwfr_slice
  complex  , dimension(:,:,:)    , allocatable, private :: dwfs_slice, dwfr_slice
  complex  , dimension(:,:,:)    , allocatable, private :: wtmp_slice

  real     , dimension(:,:,:,:)  , allocatable, private :: imag

  real     , dimension(:)        , allocatable, private :: kx2 
  real     , dimension(:)        , allocatable, private :: ky2
  real     , dimension(:)        , allocatable, private :: tapx
  real     , dimension(:)        , allocatable, private :: tapy

  integer*8, dimension(:)        , allocatable, private :: plan_2dsF, plan_2dsI
  integer*8, dimension(:)        , allocatable, private :: plan_2drF, plan_2drI
  integer*8, dimension(:)        , allocatable, private :: plan_2dtF, plan_2dtI  

  integer  , dimension(:,:)      , allocatable, private :: index_mxs, index_mxr
  integer  , dimension(:,:)      , allocatable, private :: index_mys, index_myr

  integer  , dimension(:,:,:)    , allocatable, private :: iwrk2d
  integer  , dimension(:,:,:)    , allocatable, private :: ivwrk2d, iqwrk2d
  real     , dimension(:,:)      , allocatable, private :: rwrk1d
  complex  , dimension(:,:)      , allocatable, private :: cwrk1d
  complex  , dimension(:,:,:,:)  , allocatable, private :: cwrk3d

  integer                                     , private :: nzrept
  integer                                     , private :: bscnop
  character(len=6)                            , private :: bsccoe
  character(len=6)                            , private :: bsctyp
 
  logical                                     , private :: wntwfs
  logical                                     , private :: wntwfr
  logical                                     , private :: report
  logical                                     , private :: memchk

  integer                                     , private :: imgfwd_sign
  logical                                     , private :: imgfwd_oper

  integer                                     , private :: tombwd_sign
  logical                                     , private :: tombwd_oper

  integer                                     , private :: ompchunk, nth
  integer  ,                          external, private :: omp_get_num_procs
  integer  ,                          external, private :: omp_get_num_threads
  integer  ,                          external, private :: omp_set_num_threads
  integer  ,                          external, private :: omp_get_thread_num

  real     ,                         parameter, private :: PI = 3.14159265359

contains


  integer function bwi_kernel_init(dgrid_out, pgrid_out, vgrid_out, qgrid_out, igrid_out, hgrid_out, ftag_out_) result(stat)
    type(dat_grid)   :: dgrid_out
    type(pro_grid)   :: pgrid_out
    type(vel_grid)   :: vgrid_out
    type(q_grid)     :: qgrid_out
    type(img_grid)   :: igrid_out
    type(hes_grid)   :: hgrid_out
    type(bwi_ftag)   :: ftag_out_

    real             :: bndpr, bndpi
    character(len=7) :: imgfwd_type
    character(len=7) :: tombwd_type

integer :: i

    ftag_%csou_ = "csou"       !!.. source function
    ftag_%crec_ = "crec"       !!.. receiver data
    ftag_%bvel_ = "bvel"       !!.. background velocity
!    ftag_%dvel_ = "dvel"       !!.. velocity perturbation 
    ftag_%bq_   = "bq"         !!.. background q
    ftag_%dq_   = "dq"         !!.. q perturbation 
    ftag_%bimg_ = "bimg"       !!.. bakground image
    ftag_%dimg_ = "dimg"       !!.. image perturbation
    ftag_%grdo_ = "grdo"       !!.. gradient at current iteration
    ftag_%grdp_ = "grdp"       !!.. gradient at next iteration 
    ftag_%srch_ = "srch"       !!.. search direction 
    ftag_%bwfs_ = "bwfs"       !!.. source wavefield
    ftag_%bwfr_ = "bwfr"       !!.. receiver wavefield
    ftag_%sfnc_ = "sfnc"       !!.. source signature
    ftag_%gsou_ = "gsou"       !!.. source function for source-side Green's function
    ftag_%grec_ = "grec"       !!.. source function for receiver-side Green's function
    ftag_%hess_ = "hess"       !!.. Hessian
    ftag_%imsk_ = "imsk"       !!.. image mask
    ftag_%gmsk_ = "gmsk"       !!.. gradient mask
    ftag_%hsou_ = "hsou"       !!.. source header

!   read parameters
!   ---------------
    call from_param("radian", dgrid%radian, .true.  ) !!.. whether the input frequency is in radian or not
    call from_param("offset", dgrid%offset, .false. ) !!.. whether input data are in receiver or offset coordinates
    call from_param("comsou", dgrid%comsou, .false. ) !!.. point source or composite source
    call from_param("hdrsou", dgrid%hdrsou, .false. ) !!.. input source header or not
    call from_param("smtsou", dgrid%smtsou, .false. ) !!.. smooth point source or not
    call from_param("cpximg", igrid%cpximg, .false. ) !!.. output complex image
    call from_param("stkimg", igrid%stkimg, .true.  ) !!.. output image stacked over shot axis
    call from_param("imgmsk", igrid%imgmsk, .false. ) !!.. whether or not to apply image mask
    call from_param("grdmsk", qgrid%grdmsk, .false. ) !!.. whether or not to apply gradient maskt
    call from_param("srcbnd", pgrid%srcbnd, .false. ) !!.. boundary condition for source function, change dipole source into monopole
    call from_param("recbnd", pgrid%recbnd, .false. ) !!.. boundary condition for the receivers. use this only for impluse response test 
    call from_param("vsfc"  , pgrid%vsfc  , 1500.   ) !!.. surface velocity for boundary conditions
    call from_param("ntpx"  , pgrid%ntpx  , 1       ) !!.. taper width in x
    call from_param("ntpy"  , pgrid%ntpy  , 1       ) !!.. taper width in y
    call from_param("nvrf"  , pgrid%nvrf  , 1       ) !!.. number of reference velocities
    call from_param("nqrf"  , pgrid%nqrf  , 1       ) !!.. number of reference q
    call from_param("oper"  , pgrid%oper  , "ssfpi" ) !!.. extrapolator type, phsft/ssfou/impfd/foufd/phspi/[ssfpi]
    call from_param("mode"  , pgrid%mode  , "imgadj") !!.. mode to run
    call from_param("wnthes", hgrid%wnthes, .false. ) !!.. whether or not to output Hessian
    call from_param("difsou", hgrid%difsou, .false. ) !!.. whether or not to use the same source function for both migration and Hessian
    call from_param("grecin", hgrid%grecin, .false. ) !!.. whether or not to input source functions for receiver-side Green's function
    call from_param("hestyp", hgrid%hestyp, 1       ) !!.. Hessian type, 1 for source wavefield intensity; 2 for phase-encoded Hessian
    call from_param("rdmtyp", hgrid%rdmtyp, 2       ) !!.. random phase encoding type, 1 for no encoding; 2 for phase code; 3 for binary code; 4 for random time delay
    call from_param("nws"   , wblock%ns   , 1       ) !!.. number of frequencies to hold in memory
    call from_param("bscnop", bscnop      , 2       ) !!.. number of terms for Born scattering operator
    call from_param("bsctyp", bsctyp      , "fourie") !!.. type of Born scattering, "fourie" for Fourier domain; "tayexp" for Taylor expansion 
    call from_param("bsccoe", bsccoe      , "optimz") !!.. coefficients for Born scattering operator, "taylor" for Taylor coeffs; "optimz" for optimized coeffs
    call from_param("nzrept", nzrept      , 50      ) !!.. report every nzrept depths
    call from_param("wntwfs", wntwfs      , .false. ) !!.. whether or not to output source wavefield
    call from_param("wntwfr", wntwfr      , .false. ) !!.. whether or not to output receiver wavefield
    call from_param("report", report      , .true.  ) !!.. print report or not 
    call from_param("memchk", memchk      , .true.  ) !!.. memory check or not


    call from_param("imgfwd_type", imgfwd_type, "bckward") !!.. extrapolation type for imaging/modeling
    call from_param("tombwd_type", tombwd_type, "adjoint") !!.. extrapolation type for tomography



    select case (imgfwd_type)
      case ("bckward")
           imgfwd_oper = .false.; imgfwd_sign = +1
      case ("adjoint")
           imgfwd_oper = .true. ; imgfwd_sign = -1
      case default
           write(0,*)"imgfwd_type has to be either <bckward> or <adjoint>"
           STOP
    end select

    select case (tombwd_type)
      case ("bckward")
           tombwd_oper = .false.; tombwd_sign = +1
      case ("adjoint")
           tombwd_oper = .true. ; tombwd_sign = -1
      case default
           write(0,*)"tombwd_type has to be either <bckward> or <adjoint>"
           STOP
    end select
    
    if (pgrid%oper=="phspi" .or. pgrid%oper=="ssfpi") then
       pgrid%pdip  = 0
       pgrid%trick = 0.
       pgrid%bndp  = 0.

    else 
       call from_param("pdip"  , pgrid%pdip  , 45     ) !!.. propagation dip for implicit finite difference
       call from_param("trick" , pgrid%trick , 0.14   ) !!.. 1/6 trick
       call from_param("bndpr" , bndpr       , 0.     ) !!.. boundary condition, real part
       call from_param("bndpi" , bndpi       , 0.     ) !!.. boundary condition, imaginary part
       call from_param("bullet", pgrid%bullet, .false.) !!.. use bulletproof scheme or not
       call from_param("errflt", pgrid%errflt, .true. ) !!.. apply Li's compensation filter or not
       pgrid%bndp = cmplx(bndpr, bndpi)
    end if

    if (report) then
       write(0,*)"imgfwd_type = ", imgfwd_type
       write(0,*)"imgfwd_oper = ", imgfwd_oper, imgfwd_sign
       write(0,*)"tombwd_type = ", tombwd_type
       write(0,*)"tombwd_oper = ", tombwd_oper, tombwd_sign
       write(0,*)"oper = ", pgrid%oper, " nvrf = ", pgrid%nvrf," nqrf = ", pgrid%nqrf
    end if

    select case (pgrid%mode)
      case ("imgfwd")
           if (report) then 
              write(0,*) "Imaging: FORWARD"
           end if 
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%bq_)
           call auxinout(ftag_%bimg_)
           call bwi_histos_imgfwd(dgrid, pgrid, vgrid, qgrid, igrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_,ftag_%bq_, ftag_%bimg_)
      case ("imgadj")
           if (report) then
              write(0,*) "Imaging: ADJOINT"
           end if 
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%bq_)
           call auxinout(ftag_%bimg_)
           call bwi_histos_imgadj(dgrid, pgrid, vgrid, qgrid, igrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_,ftag_%bq_, ftag_%bimg_, ftag_%hsou_)

      case ("imginv")
      case ("tomfwd")
           if (report) then
              write(0,*) "Tomography: FORWARD"
           end if
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
         !  call auxinout(ftag_%dvel_)
           call auxinout(ftag_%bq_)
           call auxinout(ftag_%dq_)
           call auxinout(ftag_%dimg_)
           call bwi_histos_tomfwd(dgrid, pgrid, vgrid, qgrid,igrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bq_, ftag_%dq_, ftag_%dimg_)

      case ("tomadj") 
           if (report) then
               write(0,*) "Tomography: ADJOINT" 
           end if
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
   !        call auxinout(ftag_%dvel_)
           call auxinout(ftag_%bq_)
           call auxinout(ftag_%dq_)
           call auxinout(ftag_%dimg_)
           call bwi_histos_tomadj(dgrid, pgrid, vgrid,qgrid, igrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bq_, ftag_%dq_, ftag_%dimg_)

      case ("tominv")
           if (report) then
              write(0,*) "Tomography: INVERSION"
           end if
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%bq_)
           call auxinout(ftag_%bimg_)
           call auxinout(ftag_%grdo_)
           call auxinout(ftag_%grdp_)
           call auxinout(ftag_%srch_)
           if (hgrid%wnthes) then
              call auxinout(ftag_%hess_)
              if (hgrid%difsou) then
                 call auxinout(ftag_%gsou_)
              else
                 ftag_%gsou_ = ftag_%csou_
              end if
              if (hgrid%grecin   ) call auxinout(ftag_%grec_)
              if (hgrid%hestyp==2) call auxinout(ftag_%sfnc_)
           end if
           if (igrid%imgmsk) call auxinout(ftag_%imsk_)
           if (qgrid%grdmsk) call auxinout(ftag_%gmsk_)
           call bwi_histos_tominv(dgrid, pgrid, vgrid, qgrid, igrid, hgrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_,ftag_%bq_, &
                                  ftag_%bimg_, ftag_%hess_, ftag_%grdo_, ftag_%grdp_, ftag_%srch_) 

      case ("hesian")
           if (report) then
               write(0,*) "HESSIAN"
           end if 
           call auxinout(ftag_%gsou_)
           call auxinout(ftag_%hess_)
           call auxinout(ftag_%bvel_)
           if (hgrid%grecin   ) call auxinout(ftag_%grec_)
           if (hgrid%hestyp==2) call auxinout(ftag_%sfnc_)
           call bwi_histos_hesian(dgrid, pgrid, vgrid, qgrid, igrid, hgrid, ftag_%sfnc_, ftag_%gsou_, ftag_%grec_, ftag_%bvel_, ftag_%bq_, ftag_%hess_)

    end select

!   setup OpenMP stuff
!   ------------------
    ompchunk = 1
    nth      = 1

    !$OMP PARALLEL
    if (omp_get_thread_num() == 0) then
       nth = omp_get_num_threads()
       if (report) then 
          write(0,*)"nthread =", nth, omp_get_num_procs()
       end if
    end if
    !$OMP END PARALLEL

!   radian
!   ------
    if (.not.dgrid%radian) then
       dgrid%aw%d = dgrid%aw%d*2*PI
       dgrid%aw%o = dgrid%aw%o*2*PI
    end if





!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!   if present the source header, read it
!   -------------------------------------
if (dgrid%hdrsou) then
   allocate(dgrid%hsou(2,dgrid%nshg))
   call sep_read(dgrid%hsou, ftag_%hsou_)
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!   calculate the global parameters for the first shot
!   --------------------------------------------------
    call bwi_glocal(pgrid, ogrid, vgrid, qgrid, dgrid, igrid, hgrid, 1)


!   calculate the frequency blocks
!   ------------------------------
    call base_blkcal(wblock, dgrid%aw)

!   check memory
!   ------------
    call bwi_memory_chck(memchk)

!   allocate arrays
!   ---------------
    call bwi_arrays_init()

!   initialize modules
!   ------------------
    call bwi_module_init()
    if (wntwfs) then
       call auxinout(ftag_%bwfs_)
       call bwi_histos_wfld(dgrid, pgrid, ftag_%bwfs_)
    end if
    if (wntwfr) then
       call auxinout(ftag_%bwfr_)
       call bwi_histos_wfld(dgrid, pgrid, ftag_%bwfr_)
    end if

    dgrid_out = dgrid
    pgrid_out = pgrid
    vgrid_out = vgrid   !! bug danger
    igrid_out = igrid
    hgrid_out = hgrid
    qgrid_out = qgrid  
    ftag_out_ = ftag_
    stat = 0
  end function

  integer function bwi_kernel_clos() result(stat)
    stat = 0
  end function

  integer function bwi_kerimg_oper(adj, add, qcomp, csou_, crec_, bvel_,bq_, bimg_, qcomp_clip) result(stat)
    logical           :: adj, add
    character(len=*)  :: csou_, crec_, bvel_, bq_, bimg_

    integer           :: ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak
    integer           :: ish, ishx, ishy, izs

    logical	      :: qcomp
    real	      :: qcomp_clip

    if (.not.add) then
       if (adj) then
          if (igrid%cpximg) then
             call bwi_fio_cimg_init(igrid, bimg_)
          else
             call bwi_fio_rimg_init(igrid, bimg_)
          end if
       else
          call bwi_fio_crec_init(dgrid, crec_)
       end if
    end if


!   loop over shots
!   ---------------
    do ish=1, dgrid%nshg
!      output image coordinates
!      ------------------------
       if (igrid%stkimg) then
          ishx = 1
          ishy = 1
       else
          ishx = mod((ish-1)/1          , dgrid%asx%n)+1
          ishy = mod((ish-1)/dgrid%asx%n, dgrid%asy%n)+1
       end if
!      some backups
!      ------------
       ogrid_nx_bak = ogrid%ax%n; ogrid_ny_bak = ogrid%ay%n; ogrid_nz_bak = ogrid%az%n
!      calculate global parameters
!      ---------------------------
       call bwi_glocal(pgrid, ogrid, vgrid, qgrid, dgrid, igrid, hgrid, ish)
       if (igrid%ax%n>1 .and. ogrid%ax%n==0) cycle
       if (igrid%ay%n>1 .and. ogrid%ay%n==0) cycle
!      array safe guarding
!      -------------------
       call bwi_arrays_rset(ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak)

       if (report) call bwi_report(dgrid, vgrid, qgrid, pgrid, ogrid, igrid, ish)

!      initialize extrapolator
!      -----------------------
       stat = bwi_wfdexp_init(kx2,ky2,pgrid%ax%n,pgrid%ay%n,pgrid%ax%d,pgrid%ay%d,pgrid%az%d,pgrid%pdip, &
                              pgrid%nvrf,pgrid%nqrf,pgrid%oper,pgrid%trick,pgrid%bndp,pgrid%bullet,pgrid%errflt)

!      reference velocity
!      ------------------ 
       stat = bwi_velref_init(pgrid%nvrf)

!      reference q
!      ------------------ 
       stat = bwi_qref_init(pgrid%nqrf)

!      initialize imaging condition
!      ----------------------------
       stat = bwi_imgcon_init(ogrid, pgrid, dgrid%aw%n, dgrid%nshg, index_mxs, index_mxr, index_mys, index_myr)

!      compute for 1 shot
!      ------------------
       call bwi_img1st(adj, add, qcomp, csou_, crec_, bvel_, bq_, bimg_, ish, ishx, ishy, qcomp_clip)


!      release some memory
!      -------------------
       stat = bwi_imgcon_clos()
       stat = bwi_velref_clos()
       stat = bwi_wfdexp_clos()
    end do
    stat = 0
  end function

! imaging for one shot
  subroutine bwi_img1st(adj, add, qcomp, csou_, crec_, bvel_, bq_, bimg_, ish, ishx, ishy, qcomp_clip)
    logical          :: adj, add
    character(len=*) :: csou_, crec_, bvel_, bq_, bimg_
    integer          :: ish, ishx, ishy
    type(axis)       :: aws
    integer          :: iws, iwb, izs, jzs, ith, stat
    real             :: ws, vsfc

    logical	     :: qcomp
    real	     :: qcomp_clip
    real	     :: p1, p2, p3, p4, t1,t2
!   shot-profile migration 
!   ----------------------
    if (adj) then
!      loop over frequency block
!      -------------------------
       do iwb=1, wblock%nb
          if (report) write(0,*)"Working at iwb = ", iwb, "out of ", wblock%nb
          aws%o = wblock%os(iwb); aws%d = wblock%ds; aws%n = wblock%blen(iwb)
          bwfs = 0.; bwfr = 0.
!         read source and receiver wavefields
!         -----------------------------------
          call bwi_fio_csou(csou, dgrid, wblock%indx(iwb), aws%n, ish, csou_, "reed")
          call bwi_fio_crec(crec, dgrid, wblock%indx(iwb), aws%n, ish, crec_, "reed")
          call bwi_wfs_init(.true., .false., bwfs(:,:,:,1), csou, dgrid, pgrid)
          call bwi_wfr_init(.true., .false., bwfr(:,:,:,1), crec, dgrid, pgrid)

!         apply surface boundary condition for source (change dipole source into mono-pole)
!         ---------------------------------------------------------------------------------
          if (pgrid%srcbnd) then
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,ith,ws) 
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,1)
                call bwi_sbcond(dwfs_slice(:,:,ith), kx2, ky2, pgrid%vsfc, ws, pgrid%ax%n, pgrid%ay%n, plan_2dsF(ith), plan_2dsI(ith))
                bwfs(:,:,iws,1) = dwfs_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end if

!         loop over depth
!         ----------------
          do izs=1, pgrid%nzstep
!            read velocity model
!            -------------------
             if (izs>1) then
                call bwi_fio_velo(bvel, vgrid, pgrid, izs-1, bvel_, "reed")
                stat = bwi_velref_oper(bvel, bvrf)
             end if
!            read q model
!            -------------------
             if (izs>1) then
                call bwi_fio_q(bq, qgrid, pgrid, izs-1, bq_, "reed")
                stat = bwi_qref_oper(bq, bqrf)
             end if


!            initialize arrays
!            -----------------
             if (igrid%cpximg) then
                cimg = 0. 
             else
                rimg = 0.
             end if
!            read images
!            -----------
             if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                if (igrid%cpximg) then
                   call bwi_fio_cimg(cimg(:,:,:,:,1), igrid, ogrid, jzs, ishx, ishy, bimg_, "reed")
                else
                   call bwi_fio_rimg(rimg(:,:,:,:,1), igrid, ogrid, jzs, ishx, ishy, bimg_, "reed")
                end if
             end if
!            initialize some arrays
!            ----------------------
             dwfs_slice = 0.; dwfr_slice = 0.
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                if (izs==1) then
                   dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs)
                   dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs)
                else
                   dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs-1)
                   dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs-1)
                   stat = bwi_tapers_oper(dwfs_slice(:,:,ith),pgrid%ax%n,pgrid%ay%n)
                   stat = bwi_tapers_oper(dwfr_slice(:,:,ith),pgrid%ax%n,pgrid%ay%n)
                   stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfs_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith),qcomp_clip, p1, p2, p3, p4)

                   stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfr_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith),qcomp_clip, p1, p2, p3, p4)
                   bwfs(:,:,iws,izs) = dwfs_slice(:,:,ith)
                   bwfr(:,:,iws,izs) = dwfr_slice(:,:,ith)
                end if
                if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                   if (igrid%cpximg) then
                      stat = bwi_imgcon_oper(.false., cimg(:,:,:,:,ith), dwfs_slice(:,:,ith), dwfr_slice(:,:,ith), ws, "all")
                   else
                      stat = bwi_imgcon_oper(.false., rimg(:,:,:,:,ith), dwfs_slice(:,:,ith), dwfr_slice(:,:,ith), ws, "all")
		     
                   end if
                end if
             end do
             !$OMP END PARALLEL DO
!            write out images
!            ----------------
             if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                if (igrid%cpximg) then
                   call bwi_fio_cimg(sum(cimg,5), igrid, ogrid, jzs, ishx, ishy, bimg_, "rite")
                else
                   call bwi_fio_rimg(sum(rimg,5), igrid, ogrid, jzs, ishx, ishy, bimg_, "rite")
                end if
             end if
          end do
!	write(0,*)"bwfs, bwfr",sum(bwfs)/size(bwfs),size(bwfs),sum(bwfr)/size(bwfr),size(bwfr)
!         output source wavefield if requested
!         ------------------------------------
          if (wntwfs) call bwi_fio_wfld(conjg(bwfs), dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfs_, "rite")
!         output receiver wavefield if requested
!         --------------------------------------  
          if (wntwfr) call bwi_fio_wfld(      bwfr , dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfr_, "rite")
       end do

    else

!      frequency block
!      --------------- 
       do iwb=1, wblock%nb
          if (report) write(0,*)"Working at iwb = ", iwb, "out of ", wblock%nb
          aws%o = wblock%os(iwb); aws%d = wblock%ds; aws%n = wblock%blen(iwb)


          bwfs = 0.; bwfr = 0.
!         read source wavefield
!         ---------------------
          call bwi_fio_csou(csou, dgrid, wblock%indx(iwb), aws%n, ish, csou_, "reed")
          call bwi_wfs_init(.true., .false., bwfs(:,:,:,1), csou, dgrid, pgrid)

!         apply surface boundary condition for source (change dipole source into mono-pole)
!         ---------------------------------------------------------------------------------
          if (pgrid%srcbnd) then
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,ith,ws) 
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,1)
                call bwi_sbcond(dwfs_slice(:,:,ith), kx2, ky2, pgrid%vsfc, ws, pgrid%ax%n, pgrid%ay%n, plan_2dsF(ith), plan_2dsI(ith))
                bwfs(:,:,iws,1) = dwfs_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end if

!         propagate source wavefield
!         --------------------------
          do izs=2, pgrid%nzstep
!            read velocity model
!            -------------------
             call bwi_fio_velo(bvel, vgrid, pgrid, izs-1, bvel_, "reed")
             stat = bwi_velref_oper(bvel, bvrf)
!            read q model
!            -------------------
             call bwi_fio_q(bq, qgrid, pgrid, izs-1, bq_, "reed")
             stat = bwi_qref_oper(bq, bqrf)
             dwfs_slice = 0.

             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs-1)
                stat = bwi_tapers_oper(dwfs_slice(:,:,ith),pgrid%ax%n,pgrid%ay%n)
                stat = bwi_wfdexp_oper(.false.,-1, .false., dwfs_slice(:,:,ith),bvel,bq, bvrf, bqrf, ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith), qcomp_clip, p1, p2, p3, p4)
                bwfs(:,:,iws,izs) = dwfs_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end do
!         loop over depth
!         ---------------
!	write(0,*)"bwfs",sum(dwfs_slice)
          do izs=pgrid%nzstep, 1, -1
!            read velocity model
!            -------------------
             call bwi_fio_velo(bvel, vgrid, pgrid, izs, bvel_, "reed")
             stat = bwi_velref_oper(bvel, bvrf)

!            read q model
!            -------------------
             call bwi_fio_q(bq, qgrid, pgrid, izs, bq_, "reed")
             stat = bwi_qref_oper(bq, bqrf)


!            initialize arrays
!            -----------------
             if (igrid%cpximg) then 
                cimg = 0. 
             else
                rimg = 0.
             end if
!            read images
!            -----------
             if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                if (igrid%cpximg) then
                   call bwi_fio_cimg(cimg(:,:,:,:,1), igrid, ogrid, jzs, ishx, ishy, bimg_, "reed")
                   do ith=1, nth
                      cimg(:,:,:,:,ith) = cimg(:,:,:,:,1)
                   end do
                else
                   call bwi_fio_rimg(rimg(:,:,:,:,1), igrid, ogrid, jzs, ishx, ishy, bimg_, "reed")
                   do ith=1, nth
                      rimg(:,:,:,:,ith) = rimg(:,:,:,:,1)
                   end do
                end if
             end if
!            initialize some arrays
!            ----------------------
             dwfs_slice = 0.; dwfr_slice = 0.
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs)
                if (izs==pgrid%nzstep) then
                   dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs)
                else
                   dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs+1)
                   stat = bwi_wfdexp_oper(imgfwd_oper,imgfwd_sign,.false., dwfr_slice(:,:,ith),bvel,bq, bvrf, bqrf, ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith),qcomp_clip, p1, p2, p3, p4)
                   stat = bwi_tapers_oper(dwfr_slice(:,:,ith),pgrid%ax%n,pgrid%ay%n)
                end if
                if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                   if (igrid%cpximg) then
                      stat = bwi_imgcon_oper(.true., cimg(:,:,:,:,ith), dwfs_slice(:,:,ith), dwfr_slice(:,:,ith), ws, "rec")
                   else
                      stat = bwi_imgcon_oper(.true., rimg(:,:,:,:,ith), dwfs_slice(:,:,ith), dwfr_slice(:,:,ith), ws, "rec")
                   end if
                end if
                bwfr(:,:,iws,izs) = dwfr_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO

          end do
!         write out modeled data
!         ----------------------

          if (add) call bwi_fio_crec(crec, dgrid, wblock%indx(iwb), aws%n, ish, crec_, "reed")
          call bwi_wfr_init(.false., add, bwfr(:,:,:,1), crec, dgrid, pgrid)
          call bwi_fio_crec(crec, dgrid, wblock%indx(iwb), aws%n, ish, crec_, "rite")
!         output source wavefield, if requested
!         -------------------------------------
          if (wntwfs) call bwi_fio_wfld(conjg(bwfs), dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfs_, "rite")
!         output receiver wavefield, if requested
!         ---------------------------------------
          if (wntwfr) call bwi_fio_wfld(      bwfr , dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfr_, "rite")
       end do
    end if
    return
  end subroutine


!!=========================================================================================================================================
! 			tomography-kenel 
!!=========================================================================================================================================

  integer function bwi_kertom_oper(adj, add, qcomp, csou_, crec_, bvel_,bq_, dq_, dimg_, qcomp_clip) result(stat)
    logical           :: adj, add
    character(len=*)  :: csou_, crec_, bvel_,  dimg_, bq_, dq_

    type(axis)        :: aws
    integer           :: izs, ish, ishx, ishy
    integer           :: ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak

    logical	      :: qcomp
    real	      :: qcomp_clip

    if (.not.add) then
       if (adj) then
    !      call bwi_fio_dvel_init(vgrid, dvel_)
          call bwi_fio_dq_init(qgrid, dq_)
       else
          call bwi_fio_rimg_init(igrid, dimg_)
       end if
    end if

!   loop over shots
!   ---------------
    do ish=1, dgrid%nshg
!      output image coordinates
!      ------------------------
       if (igrid%stkimg) then
          ishx = 1
          ishy = 1
       else
          ishx = mod((ish-1)/1          , dgrid%asx%n)+1
          ishy = mod((ish-1)/dgrid%asx%n, dgrid%asy%n)+1
       end if
!      some backups
!      ------------
       ogrid_nx_bak = ogrid%ax%n; ogrid_ny_bak = ogrid%ay%n; ogrid_nz_bak = ogrid%az%n
!      calculate the global parameters
!      -------------------------------
       call bwi_glocal(pgrid, ogrid, vgrid, qgrid, dgrid, igrid, hgrid, ish)
       if (igrid%ax%n>1.and.ogrid%ax%n == 0) cycle
       if (igrid%ay%n>1.and.ogrid%ay%n == 0) cycle
       call bwi_arrays_rset(ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak)
       if (report) call bwi_report(dgrid, vgrid, qgrid, pgrid, ogrid, igrid, ish)
!      initialize extrapolators
!      ------------------------
       stat = bwi_wfdexp_init(kx2,ky2,pgrid%ax%n,pgrid%ay%n,pgrid%ax%d,pgrid%ay%d,pgrid%az%d,pgrid%pdip,&
                              pgrid%nvrf,pgrid%nqrf, pgrid%oper,pgrid%trick,pgrid%bndp,pgrid%bullet,pgrid%errflt)
!      initialize reference velocity
!      -----------------------------
       stat = bwi_velref_init(pgrid%nvrf)
!      initialize reference q
!      -----------------------------
       stat = bwi_qref_init(pgrid%nqrf)
!      initialize imaging condition
!      ----------------------------
       stat = bwi_imgcon_init(ogrid, pgrid, dgrid%aw%n, dgrid%nshg, index_mxs, index_mxr, index_mys, index_myr)
!      initialize the Born scattering operator
!      ---------------------------------------
       stat = bwi_bornsc_init(kx2,ky2,pgrid%ax%d,pgrid%ay%d,pgrid%az%d,pgrid%ax%n,pgrid%ay%n,pgrid%nvrf,pgrid%nqrf,bscnop,bsctyp,bsccoe)

!stat = bwi_bornsc_init(kx2,ky2,pgrid%ax%d,pgrid%ay%d,pgrid%az%d,pgrid%ax%n,pgrid%ay%n,bscnop,2)
!      apply the tomographic operator
!      ------------------------------
       call bwi_tom1st(adj, add, qcomp, csou_, crec_, bvel_, bq_, dq_, dimg_, ish, ishx, ishy, qcomp_clip)
!      close subroutines
!      -----------------
       stat = bwi_bornsc_clos()
       stat = bwi_imgcon_clos()
       stat = bwi_velref_clos()
       stat = bwi_qref_clos()
       stat = bwi_wfdexp_clos() 
    end do
    stat = 0
  end function



!!=========================================================================================================================================
! 			tomography-subroutine -not used temporarily
!!=========================================================================================================================================
  subroutine bwi_tom1st(adj, add, qcomp, csou_, crec_, bvel_, bq_, dq_, dimg_, ish, ishx, ishy, qcomp_clip)
    logical          :: adj, add
    character(len=*) :: csou_, crec_, bvel_,  bq_, dq_,dimg_
    integer          :: ish, ishx, ishy 

    type(axis)       :: aws
    integer          :: iwb, iws, izs, jzs, ith, stat
    real             :: ws

    logical	      :: qcomp
    real	      :: qcomp_clip
    real	      ::p1, p2, p3, p4

    if (adj) then     !!.. dQ = F' dI
       do iwb=1, wblock%nb
          if (report) write(0,*)"Working at iwb = ", iwb, "out of ", wblock%nb
          aws%o = wblock%os(iwb); aws%d = wblock%ds; aws%n = wblock%blen(iwb)
          bwfs = 0.; bwfr = 0.
          dwfs = 0.; dwfr = 0.
          call bwi_fio_csou(csou, dgrid, wblock%indx(iwb), aws%n, ish, csou_, "reed")
          call bwi_fio_crec(crec, dgrid, wblock%indx(iwb), aws%n, ish, crec_, "reed")
          call bwi_wfs_init(.true., .false., bwfs(:,:,:,1), csou, dgrid, pgrid)
          call bwi_wfr_init(.true., .false., bwfr(:,:,:,1), crec, dgrid, pgrid)

!         apply surface boundary condition for source (change dipole source into mono-pole)
!         ---------------------------------------------------------------------------------
          if (pgrid%srcbnd) then
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,ith,ws) 
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,1)
                call bwi_sbcond(dwfs_slice(:,:,ith), kx2, ky2, pgrid%vsfc, ws, pgrid%ax%n, pgrid%ay%n, plan_2dsF(ith), plan_2dsI(ith))
                bwfs(:,:,iws,1) = dwfs_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end if

!         compute the background wavefields
!         --------------------------------- 
          do izs=1, pgrid%nzstep-1
             call bwi_fio_velo(bvel, vgrid, pgrid, izs, bvel_, "reed")
             call bwi_fio_q(bq, qgrid, pgrid, izs, bq_, "reed")
             stat = bwi_velref_oper(bvel, bvrf)
             stat = bwi_qref_oper(bq, bqrf)
             dwfs_slice = 0.; dwfr_slice = 0.
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs)
                dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs)
                stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfs_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith),qcomp_clip, p1, p2, p3, p4)
                stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfr_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith),qcomp_clip, p1, p2, p3, p4)

                stat = bwi_tapers_oper(dwfs_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
                stat = bwi_tapers_oper(dwfr_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
                bwfs(:,:,iws,izs+1) = dwfs_slice(:,:,ith)
                bwfr(:,:,iws,izs+1) = dwfr_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end do

          !!.. dP_i = E_{i}'(dP_{i-1}+G_{i-1}dQ_{i-1})
          do izs=pgrid%nzstep-1, 1, -1
!            read background velocity at current depth step: bV_{i} 
!            ------------------------------------------------------
             call bwi_fio_velo(bvel, vgrid, pgrid, izs, bvel_, "reed")
!            read background Q at current depth step: bQ_{i} 
             call bwi_fio_q(bq, qgrid, pgrid, izs, bq_, "reed")
!            calculate reference velocity
!            ----------------------------
             stat = bwi_velref_oper(bvel, bvrf)
!            calculate reference Q
!            ----------------------------
             stat = bwi_qref_oper(bq, bqrf)
!            initialize arrays
!            -----------------
             rimg = 0.;dq =0. 
!            read the perturbed image at previous depth level: dI_{i+1}
!            ----------------------------------------------------------
             if (izs+1>=ogrid%index_outpro_zmin .and. izs+1<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                call bwi_fio_rimg(rimg(:,:,:,:,1), igrid, ogrid, jzs+1, ishx, ishy, dimg_, "reed")
             end if
!            copy it to all threads
!            ----------------------
             do ith=1, nth
                rimg(:,:,:,:,ith) = rimg(:,:,:,:,1)
             end do
!            read the existing perturbed velocity at current depth level for accumulating dV_{i}
!            -----------------------------------------------------------------------------------
 !            call bwi_fio_velo(dvel(:,:,1), vgrid, pgrid, izs, dvel_, "reed")
!            read the existing perturbed Q at current depth level for accumulating dV_{i}
!            -----------------------------------------------------------------------------------
             call bwi_fio_q(dq(:,:,1), qgrid, pgrid, izs, dq_, "reed")
!            initialize buffers
!            ------------------
             bwfs_slice = 0.; bwfr_slice = 0.
             dwfs_slice = 0.; dwfr_slice = 0.
             wtmp_slice = 0.
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d

!               ************************ 
!               SOURCE-PART PERTURBATION
!               ************************
!               background receiver wavefield at previous depth level for adjoint imaging condition: bU_{i+1}
!               ---------------------------------------------------------------------------------------------
                bwfr_slice(:,:,ith) = bwfr(:,:,iws,izs+1)
!               perturbed auxiliary wavefield at previous depth level: dP_{i+1}
!               ----------------------------------------------------------------------
                dwfs_slice(:,:,ith) = dwfs(:,:,iws,izs+1)
!               accumulate the perturbed auxiliary wavefield with the perturbed source wavefield: dP_{i+1} = dP_{i+1} + bU_{i+1}dI_{i+1}
!               ------------------------------------------------------------------------------------------------------------------------
                if (izs+1>=ogrid%index_outpro_zmin .and. izs+1<=ogrid%index_outpro_zmax) then
                   stat = bwi_imgcon_oper(.true., rimg(:,:,:,:,ith), dwfs_slice(:,:,ith), bwfr_slice(:,:,ith), ws, "sou")
                end if
!               taper the perturbed auxiliary wavefield
!               ---------------------------------------
                stat = bwi_tapers_oper(dwfs_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
!               propagate the perturbed auxiliary wavefield to current depth level: dP_{i} = E_{i}' dP_{i+1}
!               --------------------------------------------------------------------------------------------
                stat = bwi_wfdexp_oper(tombwd_oper,tombwd_sign,qcomp, dwfs_slice(:,:,ith),bvel,bq,bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith),qcomp_clip, p1, p2, p3, p4)

!               compute the scattering operator at current depth level: G_{i}
!               -------------------------------------------------------------
                stat = bwi_bornsc_oper(bwfs_slice(:,:,ith),bwfs(:,:,iws,izs),wtmp_slice(:,:,ith), &
                                       bvel,bq,bvrf,bqrf,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith))

!stat = bwi_bornsc_oper(bwfs_slice(:,:,ith),bwfs(:,:,iws,izs),wtmp_slice(:,:,ith), &
!       bvel,ws,-1,plan_2dtF(ith),plan_2dtI(ith))


!               compute the Q perturbation at current depth level: dQ_{i} =  G_{i}' dP_{i}
!               ---------------------------------------------------------------------------------
                dq(:,:,ith) = dq(:,:,ith) + conjg(bwfs_slice(:,:,ith))*dwfs_slice(:,:,ith)
!               update the perturbed auxiliary wavefield: dP_{i}
!               ------------------------------------------------
                dwfs(:,:,iws,izs) = dwfs_slice(:,:,ith)

!               **************************
!               RECEIVER_PART PERTURBATION
!               **************************
!               background source wavefield at previous depth level for adjoint imaging condition: bD_{i+1}
!               -------------------------------------------------------------------------------------------
                bwfs_slice(:,:,ith) = bwfs(:,:,iws,izs+1)
!               perturbed auxiliary wavefield at previous depth level: dP_{i+1}
!               ---------------------------------------------------------------
                dwfr_slice(:,:,ith) = dwfr(:,:,iws,izs+1)
!               accumulate the perturbed auxiliary wavefield with the perturbed receiver wavefield: dP_{i+1} = dP_{i+1} + bD_{i+1}dI_{i+1}
!               --------------------------------------------------------------------------------------------------------------------------
                if (izs+1>=ogrid%index_outpro_zmin .and. izs+1<=ogrid%index_outpro_zmax) then
                   stat = bwi_imgcon_oper(.true., rimg(:,:,:,:,ith), bwfs_slice(:,:,ith), dwfr_slice(:,:,ith), ws, "rec")
                end if
!               taper the perturbed receiver wavefield
!               --------------------------------------
                stat = bwi_tapers_oper(dwfr_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
!               propagate the perturbed auxiliary wavefield to current depth level: dP_{i} = E{i}' dP_{i+1}
!               -------------------------------------------------------------------------------------------
                stat = bwi_wfdexp_oper(tombwd_oper,tombwd_sign,qcomp,dwfr_slice(:,:,ith),bvel,bq,bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith),qcomp_clip, p1, p2, p3, p4)

!               compute the scattering operator at current depth level: G_{i}
!               -------------------------------------------------------------
                stat = bwi_bornsc_oper(bwfr_slice(:,:,ith),bwfr(:,:,iws,izs),wtmp_slice(:,:,ith), &
                                       bvel,bq,bvrf,bqrf,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith))

!stat = bwi_bornsc_oper(bwfr_slice(:,:,ith),bwfr(:,:,iws,izs),wtmp_slice(:,:,ith), &
!       bvel,ws,-1,plan_2dtF(ith),plan_2dtI(ith))

!               compute the Q perturbation at current depth level: dV_{i} = G_{i}' dP_{i}
!               --------------------------------------------------------------------------------
                dq(:,:,ith) = dq(:,:,ith) + conjg(bwfr_slice(:,:,ith))*dwfr_slice(:,:,ith)

!               update the perturbed auxiliary wavefield: dP_{i}
!               ------------------------------------------------
                dwfr(:,:,iws,izs) = dwfr_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO

!            write out Q perturbation
!            -------------------------------
             call bwi_fio_q(sum(dq,3), qgrid, pgrid, izs, dq_, "rite")
          end do


!         output source wavefield, if requested
!         -------------------------------------
          if (wntwfs) call bwi_fio_wfld(conjg(dwfs), dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfs_, "rite")
!         output receiver wavefield, if requested
!         ---------------------------------------
          if (wntwfr) call bwi_fio_wfld(      dwfr , dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfr_, "rite")
!	write(0,*)"dwfs, dwfr",sum(dwfs)/size(dwfs),size(dwfs),sum(dwfr)/size(dwfr),size(dwfr)
       end do

    else              !!.. dI = F  dQ

       do iwb=1, wblock%nb
          if (report) write(0,*)"Working at iwb = ", iwb, "out of ", wblock%nb
          aws%o = wblock%os(iwb); aws%d = wblock%ds; aws%n = wblock%blen(iwb)
          bwfs = 0.; bwfr = 0.; dwfs = 0.; dwfr = 0.
          call bwi_fio_csou(csou, dgrid, wblock%indx(iwb), aws%n, ish, csou_, "reed")
          call bwi_fio_crec(crec, dgrid, wblock%indx(iwb), aws%n, ish, crec_, "reed")
          call bwi_wfs_init(.true., .false., bwfs(:,:,:,1), csou, dgrid, pgrid)
          call bwi_wfr_init(.true., .false., bwfr(:,:,:,1), crec, dgrid, pgrid)

!         apply surface boundary condition for source (change dipole source into mono-pole)
!         ---------------------------------------------------------------------------------
          if (pgrid%srcbnd) then
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,ith,ws) 
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,1)
                call bwi_sbcond(dwfs_slice(:,:,ith), kx2, ky2, pgrid%vsfc, ws, pgrid%ax%n, pgrid%ay%n, plan_2dsF(ith), plan_2dsI(ith))
                bwfs(:,:,iws,1) = dwfs_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end if

!         compute the backougrnd source and receiver wavefields
!         -----------------------------------------------------
          do izs=1, pgrid%nzstep-1
             call bwi_fio_velo(bvel, vgrid, pgrid, izs, bvel_, "reed")
             call bwi_fio_q(bq, qgrid, pgrid, izs, bq_, "reed")
             stat = bwi_velref_oper(bvel, bvrf)
             stat = bwi_qref_oper(bq, bqrf)
             dwfs_slice = 0.; dwfr_slice = 0.
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs)
                dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs)
                stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfs_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith),qcomp_clip, p1, p2, p3, p4)
                stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfr_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith),qcomp_clip, p1, p2, p3, p4)
                stat = bwi_tapers_oper(dwfs_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
                stat = bwi_tapers_oper(dwfr_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
                bwfs(:,:,iws,izs+1) = dwfs_slice(:,:,ith)
                bwfr(:,:,iws,izs+1) = dwfr_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end do

          !!.. dP_i = E_{i-1}(dP_{i-1}+G_{i-1}dQ_{i-1})
          do izs=2, pgrid%nzstep
!            read the background velocity at previous depth level for extrapolation bV_{i-1} => E_{i-1}
!            ------------------------------------------------------------------------------------------
             call bwi_fio_velo(bvel, vgrid, pgrid, izs-1, bvel_, "reed")
!            read the background Q at previous depth level for extrapolation bQ_{i-1} => E_{i-1}
!            ------------------------------------------------------------------------------------------
             call bwi_fio_q(bq, qgrid, pgrid, izs-1, bq_, "reed")
!            calculate the reference velocity
!            -------------------------------- 
             stat = bwi_velref_oper(bvel, bvrf)
!            calculate the reference Q
!            -------------------------------- 
             stat = bwi_qref_oper(bq, bqrf)
!            initialize some buffers
!            -----------------------
 !            rimg = 0.; dvel = 0.; dq=0.
             rimg = 0.;  dq=0.
!            read pertubed image at current depth level dI_{i}
!            -------------------------------------------------
             if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                call bwi_fio_rimg(rimg(:,:,:,:,1), igrid, ogrid, jzs, ishx, ishy, dimg_, "reed")
             end if
!            read the velocity perturbation at previous depth level dV_{i-1}
!            ---------------------------------------------------------------
    !         call bwi_fio_velo(dvel(:,:,1), vgrid, pgrid, izs-1, dvel_, "reed")      !!.. dS_{i-1}

!            read the Q perturbation at previous depth level dQ_{i-1}
!            ---------------------------------------------------------------
             call bwi_fio_q(dq(:,:,1), qgrid, pgrid, izs-1, dq_, "reed")      !!.. dQ_{i-1}
!            copy it to all threads
!            ----------------------
             do ith=1, nth
 !               dvel(:,:,ith) = dvel(:,:,1)
                dq(:,:,ith) = dq(:,:,1)
             end do
!            initilize buffers
!            -----------------
             bwfs_slice = 0.; bwfr_slice = 0.
             dwfs_slice = 0.; dwfr_slice = 0.
             wtmp_slice = 0.
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d

!               ************************
!               SOURCE-PART PERTURBATION 
!               ************************
!               background receiver wavefield at current depth level: bU_{i}
!               ------------------------------------------------------------
                bwfr_slice(:,:,ith) = bwfr(:,:,iws,izs  )
!               perturbed source wavefield at previous depth level: dD_{i-1}
!               ------------------------------------------------------------
                dwfs_slice(:,:,ith) = dwfs(:,:,iws,izs-1)
!               compute the scattering operator at previous depth step: G_{i-1}
!               ---------------------------------------------------------------
                stat = bwi_bornsc_oper(bwfs_slice(:,:,ith),bwfs(:,:,iws,izs-1),wtmp_slice(:,:,ith),&
                                       bvel,bq,bvrf,bqrf,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith))

!stat = bwi_bornsc_oper(bwfs_slice(:,:,ith),bwfs(:,:,iws,izs-1),wtmp_slice(:,:,ith),&
!       bvel,ws,-1,plan_2dtF(ith),plan_2dtI(ith))


!               accumulate the perturbed source wavefield at previous depth level: dD_{i-1} + G_{i-1}dV_{i-1}
!               ---------------------------------------------------------------------------------------------
                dwfs_slice(:,:,ith) = dwfs_slice(:,:,ith) + bwfs_slice(:,:,ith)*dq(:,:,ith) 
!               propagate the accumulated perturbed source wavefield to current depth level: dD_{i} = E_{i-1}[dD_{i-1} + G_{i-1}dV_{i-1}]
!               -------------------------------------------------------------------------------------------------------------------------
                stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfs_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith),qcomp_clip, p1, p2, p3, p4)
!               taper the perturbed source wavefield
!               ------------------------------------
                stat = bwi_tapers_oper(dwfs_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
!               correlate perturbed source wavefield with background receiver wavefield at current depth level: dI_{i} = dD_{i}*bU_{i}
!               ----------------------------------------------------------------------------------------------------------------------
                if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                   stat = bwi_imgcon_oper(.false., rimg(:,:,:,:,ith), dwfs_slice(:,:,ith), bwfr_slice(:,:,ith), ws, "sou")
                end if
!               update the wave matrix dD_{i}
!               -----------------------------
                dwfs(:,:,iws,izs) = dwfs_slice(:,:,ith)

!               **************************
!               RECEIVER-PART PERTURBATION
!               **************************
!               background source wavefield at current depth level: bD_{i}
!               ---------------------------------------------------------- 
                bwfs_slice(:,:,ith) = bwfs(:,:,iws,izs  )
!               perturbed receiver wavefield at previous depth level: dU_{i-1}
!               -------------------------------------------------------------- 
                dwfr_slice(:,:,ith) = dwfr(:,:,iws,izs-1)
!               compute the scattering operator at previous depth step: G_{i-1}
!               ---------------------------------------------------------------
                stat = bwi_bornsc_oper(bwfr_slice(:,:,ith),bwfr(:,:,iws,izs-1),wtmp_slice(:,:,ith),&
                                       bvel,bq,bvrf,bqrf,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith))

!stat = bwi_bornsc_oper(bwfr_slice(:,:,ith),bwfr(:,:,iws,izs-1),wtmp_slice(:,:,ith),&
!       bvel,ws,-1,plan_2dtF(ith),plan_2dtI(ith))

!               accumulate the perturbed receiver wavefield at previous depth level: dU_{i-1} + G_{i-1}dV_{i-1}
!               -----------------------------------------------------------------------------------------------
                dwfr_slice(:,:,ith) = dwfr_slice(:,:,ith) + bwfr_slice(:,:,ith)*dq(:,:,ith)
!               propagate the accumulated perturbed source wavefield to current depth level: dU_{i} = E_{i-1}[dU_{i-1} + G_{i-1}dV_{i-1}]
!               -------------------------------------------------------------------------------------------------------------------------
                stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfr_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), &
                                       cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith),qcomp_clip, p1, p2, p3, p4)

!               taper the perturbed receiver wavefield
!               --------------------------------------
                stat = bwi_tapers_oper(dwfr_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
!               correlate perturbed receiver wavefield with background source wavefield at current depth level: dI_{i} = dU_{i}*bD_{i}
!               ----------------------------------------------------------------------------------------------------------------------
                if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                   stat = bwi_imgcon_oper(.false., rimg(:,:,:,:,ith), bwfs_slice(:,:,ith), dwfr_slice(:,:,ith), ws, "rec")
                end if
!               update the wave matrix dU_{i}
!               ----------------------------- 
                dwfr(:,:,iws,izs) = dwfr_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO

!            write out the perturbed image
!            -----------------------------
             if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                call bwi_fio_rimg(sum(rimg,5), igrid, ogrid, jzs, ishx, ishy, dimg_, "rite")
             end if
          end do
!         output source wavefield, if requested
!         -------------------------------------
          if (wntwfs) call bwi_fio_wfld(conjg(dwfs), dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfs_, "rite")
!         output receiver wavefield, if requested
!         ---------------------------------------
          if (wntwfr) call bwi_fio_wfld(      dwfr , dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfr_, "rite")
       end do
    end if
    return
  end subroutine

!!!=========================================================================================================================================
!! 			Hessian -kenel -not used temporarily
!!!=========================================================================================================================================

   integer function bwi_kerhes_oper(add, sfnc_, gsou_, grec_, bvel_,bq_, hess_, qcomp, qcomp_clip) result(stat)
     logical          :: add
     character(len=*) :: sfnc_, gsou_, grec_, bvel_, bq_, hess_
 
     type(axis)        :: aws
     integer           :: izs, ish, ishx, ishy
     integer           :: ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak

     logical	      :: qcomp
     real	      :: qcomp_clip

 
     if (.not.add) call bwi_fio_hess_init(hgrid, hess_)
 
 !   loop over shots
 !   ---------------
     do ish=1, dgrid%nshg
 !      output image coordinates
 !      ------------------------
        if (igrid%stkimg) then
           ishx = 1
           ishy = 1
        else
           ishx = mod((ish-1)/1          , dgrid%asx%n)+1
           ishy = mod((ish-1)/dgrid%asx%n, dgrid%asy%n)+1
        end if
 !      some backups
 !      ------------
        ogrid_nx_bak = ogrid%ax%n; ogrid_ny_bak = ogrid%ay%n; ogrid_nz_bak = ogrid%az%n
 !      calculate global parameters
 !      ---------------------------
        call bwi_glocal(pgrid, ogrid, vgrid, qgrid, dgrid, igrid, hgrid, ish)
        if (igrid%ax%n>1 .and. ogrid%ax%n==0) cycle
        if (igrid%ay%n>1 .and. ogrid%ay%n==0) cycle
 !      array safe guarding
 !      -------------------
        call bwi_arrays_rset(ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak)
        if (report) call bwi_report(dgrid, vgrid, qgrid, pgrid, ogrid, igrid, ish)
 !      initialize extrapolator
 !      -----------------------
        stat = bwi_wfdexp_init(kx2,ky2,pgrid%ax%n,pgrid%ay%n,pgrid%ax%d,pgrid%ay%d,pgrid%az%d,pgrid%pdip,&
                              pgrid%nvrf,pgrid%nqrf, pgrid%oper,pgrid%trick,pgrid%bndp,pgrid%bullet,pgrid%errflt)
 !      reference velocity
 !      ------------------ 
        stat = bwi_velref_init(pgrid%nvrf)

 !      reference q
 !      ------------------ 
        stat = bwi_qref_init(pgrid%nqrf)


 !      initialize imaging condition
 !      ----------------------------
        stat = bwi_hescpt_init(ogrid, pgrid, index_mxs, index_mxr, index_mys, index_myr, hgrid%hestyp)
 !      compute for 1 shot
 !      ------------------ 
        call bwi_hes1st(add, sfnc_, gsou_, grec_, bvel_, bq_, hess_, ish, ishx, ishy, qcomp, qcomp_clip)
 !      release some memory
 !      -------------------
        stat = bwi_hescpt_clos()
        stat = bwi_velref_clos()
        stat = bwi_wfdexp_clos()
     end do
 
     stat = 0
   end function


!!!=========================================================================================================================================
!! 			Hessian -subroutine -kenel -not used temporarily
!!!=========================================================================================================================================
 subroutine bwi_hes1st(add, sfnc_, gsou_, grec_, bvel_, bq_, hess_, ish, ishx, ishy, qcomp, qcomp_clip) 
     logical          :: add
     character(len=*) :: sfnc_, gsou_, grec_, bvel_, bq_, hess_
     integer          :: ish, ishx, ishy
 
     type(axis)       :: aws
     integer          :: iwb, iws, izs, jzs, ith, stat
     real             :: ws

    logical	     :: qcomp
    real	     :: qcomp_clip

    real	     :: p1, p2, p3, p4, t1,t2
    real, dimension(:)   :: p11(nth), p22(nth), p33(nth), p44(nth),p55(nth)
 
 !   loop over frequency block
 !   -------------------------
     do iwb=1, wblock%nb
        if (report) write(0,*)"Working at iwb = ", iwb, "out of ", wblock%nb
        aws%o = wblock%os(iwb); aws%d = wblock%ds; aws%n = wblock%blen(iwb)
        bwfs = 0.
        if (hgrid%hestyp==2) bwfr = 0.
 
 !      read and initialize source wavefields
 !      -------------------------------------
        call bwi_fio_csou(csou, dgrid, wblock%indx(iwb), aws%n, ish, gsou_, "reed")
        call bwi_wfs_init(.true., .false., bwfs(:,:,:,1), csou, dgrid, pgrid)
 
 !      initialize receiver wavefield and source function
 !      -------------------------------------------------
        if (hgrid%hestyp==2) then
           call bwi_fio_sfnc(sfnc, dgrid, wblock%indx(iwb), aws%n, sfnc_, "reed")
           if (hgrid%grecin) then
              call bwi_fio_crec(crec, dgrid, wblock%indx(iwb), aws%n, ish, grec_, "reed")
		write(0,*) "I am reading grec"
           else
              stat = bwi_rdmgen_init(sfnc, aws, dgrid%ntrg, hgrid%rdmtyp)
              stat = bwi_rdmgen_oper(crec)
           end if
           call bwi_wfr_init(.true., .false., bwfr(:,:,:,1), crec, dgrid, pgrid)
        end if
 
 !      apply surface boundary condition for source (change dipole source into mono-pole)
 !      ---------------------------------------------------------------------------------
        if (pgrid%srcbnd) then
           !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,ith,ws) 
           do iws=1, aws%n
              ith = omp_get_thread_num() + 1
              ws  = aws%o + (iws-1)*aws%d
              dwfs_slice(:,:,ith) = bwfs(:,:,iws,1)
              call bwi_sbcond(dwfs_slice(:,:,ith), kx2, ky2, pgrid%vsfc, ws, pgrid%ax%n, pgrid%ay%n, plan_2dsF(ith), plan_2dsI(ith))
              bwfs(:,:,iws,1) = dwfs_slice(:,:,ith)
           end do
           !$OMP END PARALLEL DO
        end if
 
 !      loop over depth
 !      ----------------
        do izs=1, pgrid%nzstep
 !         read velocity model
 !         -------------------
           if (izs>1) then
              call bwi_fio_velo(bvel, vgrid, pgrid, izs-1, bvel_, "reed")
              stat = bwi_velref_oper(bvel, bvrf)
           end if

!          read q model
!          -------------------
           if (izs>1) then
              call bwi_fio_q(bq, qgrid, pgrid, izs-1, bq_, "reed")
              stat = bwi_qref_oper(bq, bqrf)
           end if

 !         initialize arrays
 !         -----------------
           rimg = 0.
 !         read images
 !         -----------
           if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
              jzs = izs-ogrid%index_outpro_zmin+1
              call bwi_fio_hess(rimg(:,:,:,:,1), hgrid, ogrid, jzs, ishx, ishy, hess_, "reed")
           end if
 !         initialize some arrays
 !         ----------------------
           dwfs_slice = 0.
           if (hgrid%hestyp==2) dwfr_slice = 0.
           !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
           do iws=1, aws%n
              ith = omp_get_thread_num() + 1
              ws  = aws%o + (iws-1)*aws%d
              if (izs==1) then
                 dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs)
                 if (hgrid%hestyp==2) dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs)
              else
                 dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs-1)
                 stat = bwi_tapers_oper(dwfs_slice(:,:,ith),pgrid%ax%n,pgrid%ay%n)
                    stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfs_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith),qcomp_clip, p1, p2, p3, p4)
                 bwfs(:,:,iws,izs) = dwfs_slice(:,:,ith)
                 if (hgrid%hestyp==2) then
                    dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs-1)
                    stat = bwi_tapers_oper(dwfr_slice(:,:,ith),pgrid%ax%n,pgrid%ay%n)
                    stat = bwi_wfdexp_oper(.false.,-1,qcomp,dwfr_slice(:,:,ith),bvel,bq, bvrf,bqrf,ws,ivwrk2d(:,:,ith), iqwrk2d(:,:,ith),rwrk1d(:,ith), cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith),qcomp_clip, p1, p2, p3, p4)
                    bwfr(:,:,iws,izs) = dwfr_slice(:,:,ith)
                 end if
              end if
              if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                 select case (hgrid%hestyp) 
                   case (1)
                        stat = bwi_souint_oper(rimg(:,:,:,:,ith), dwfs_slice(:,:,ith), ws)
                   case (2)
                        !stat = bwi_hesenc_oper(rimg(:,:,:,:,ith), dwfs_slice(:,:,ith), dwfr_slice(:,:,ith), sfnc(iws), ws)
 			 stat = bwi_souint_oper(rimg(:,:,:,:,ith), dwfr_slice(:,:,ith), ws)
                 end select
              end if
           end do
           !$OMP END PARALLEL DO
 !         write out images
 !         ----------------
           if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
              jzs = izs-ogrid%index_outpro_zmin+1
              call bwi_fio_hess(sum(rimg,5), hgrid, ogrid, jzs, ishx, ishy, hess_, "rite")
           end if
        end do
 !      output source wavefield if requested
 !      ------------------------------------
 write(0,*)wblock%indx(iwb), aws%n
        if (wntwfs) call bwi_fio_wfld(conjg(bwfs), dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfs_, "rite")
 !      output receiver wavefield if requested
 !      --------------------------------------  
        if (hgrid%hestyp==2 .and. wntwfr) call bwi_fio_wfld(bwfr , dgrid, pgrid, wblock%indx(iwb), aws%n, ish, ftag_%bwfr_, "rite")
 !      close random generator
 !      ----------------------
        if (.not.hgrid%grecin) stat = bwi_rdmgen_clos()
     end do
     return
   end subroutine
  
  subroutine bwi_memory_chck(memchk)
    logical          :: memchk
    double precision :: memory, s

    memory = 0
    s      = 1024. !! each term should be divided by s three times to give GB unit
!   arrays needed for all modes
!   ---------------------------
    if (pgrid%oper=="impfd" .or. pgrid%oper=="foufd") then
       memory = memory + (pgrid%ax%n*5/s+pgrid%ay%n*5/s)*nth/s*8/s + (pgrid%ax%n/s+pgrid%ay%n/s)*nth/s*4./s
    end if
    if (pgrid%oper=="phspi" .or. pgrid%oper=="ssfpi") then
       memory = memory + (pgrid%ax%n/s*pgrid%ay%n/s*2*nth/s)*8 + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*2
    end if
    memory = memory + (pgrid%ax%n/s+pgrid%ay%n/s)*4/s/s + (pgrid%ntpx/s+pgrid%ntpy/s)*4/s/s
    select case (pgrid%mode)

      case ("imgfwd")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + (pgrid%nvrf/s)*4/s/s                   &
                  + (pgrid%nqrf/s)*4/s/s       					             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*2*8                                    &
                  + (nth/s)*4*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s 

           if (igrid%cpximg) then
              memory = memory + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*8*s*s
           else
              memory = memory + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s
           end if

      case ("imgadj")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &   
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + (pgrid%nvrf/s)*4/s/s                   &
                  + (pgrid%nqrf/s)*4/s/s              					     &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*2*8                                    &
                  + (nth/s)*4*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s
           if (igrid%cpximg) then
              memory = memory + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*8*s*s
           else
              memory = memory + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s
           end if

      case ("tomfwd")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + pgrid%nvrf/s*4/s/s                     &
                  +  pgrid%nqrf/s*4/s/s         				             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*4                                      &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*5*8                                    &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s  &
                  + (nth/s)*6*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s

      case ("tomadj")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + pgrid%nvrf/s*4/s/s                     &
                  +  pgrid%nqrf/s*4/s/s           				             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*4                                      &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*5*8                                    &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s  &
                  + (nth/s)*6*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s

      case ("tominv")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + pgrid%nvrf/s*4/s/s                     &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*4                                      &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*5*8                                    &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s  &
                  + (nth/s)*6*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s      &
                  + (igrid%ahx%n/s*igrid%ahy%n/s*igrid%ax%n/s*igrid%ay%n/s)*4*s              &
                  + (vgrid%ax%n/s*vgrid%ay%n/s*vgrid%az%n/s)*3*4                       

           if (hgrid%wnthes) then
              memory = memory + (hgrid%ahx_modl%n/s*hgrid%ahy_modl%n/s*hgrid%ax_modl%n/s*hgrid%ay_modl%n/s)*4*s
           end if

           if (igrid%imgmsk) then
              memory = memory + (igrid%ax%n/s*igrid%ay%n/s)*4/s
           end if
 
      case ("hesian")
           memory = memory + wblock%ns/s*8/s/s + dgrid%ntrs/s*wblock%ns/s*8/s            &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*8*s           &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + pgrid%nvrf/s*4/s/s + pgrid%nqrf/s*4/s/s    &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*8                                  &
                  + (nth/s)*2*8/s/s                                                      &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2/s  &
                  + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s 

           if (hgrid%hestyp==2) then
              memory = memory + (dgrid%ntrg/s*wblock%ns/s)*8/s                             &
                     + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*8*s          &
                     + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*8                                 &
                     + (nth/s)*2*8/s/s                                                     &
                     + (ogrid%ahx%n/s*ogrid%ax%n/s)*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2/s 
           end if

    end select

    if (memchk) then
       write(0,*) "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
       write(0,*) "                                                           "
       write(0,*) "        Predicted memory usage: ", memory, " GB"
       write(0,*) "                                                           "
       write(0,*) "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" 
       STOP
    end if

    return
  end subroutine



  subroutine bwi_arrays_init()

!   arrays needed for all modes
!   ---------------------------
    if (pgrid%oper=="impfd" .or. pgrid%oper=="foufd") then
       allocate(cwrk1d(pgrid%ax%n*5+pgrid%ay%n*5,nth)) !!.. ifd or ffd
       allocate(rwrk1d(pgrid%ax%n+pgrid%ay%n,nth))     !!.. ifd or ffd
    end if
    if (pgrid%oper=="phspi" .or. pgrid%oper=="ssfpi") then
       allocate(cwrk3d(pgrid%ax%n,pgrid%ay%n,2,nth))   !!.. phs/ssf-epi
!      allocate(cwrk3d(pgrid%ax%n,pgrid%ay%n,2,nth))   !!.. phs/ssf-epi
       allocate(rwrk1d(pgrid%ax%n+pgrid%ay%n,nth))     !!..
       allocate(cwrk1d(pgrid%ax%n*5+pgrid%ay%n*5,nth)) !!..
    end if

    allocate(iwrk2d(pgrid%ax%n,pgrid%ay%n,nth)) !!.. phs/ssf-epi and Born PSPI 
    allocate(ivwrk2d(pgrid%ax%n,pgrid%ay%n,nth)) !!.. phs/ssf-epi and Born PSPI 
    allocate(iqwrk2d(pgrid%ax%n,pgrid%ay%n,nth)) !!.. phs/ssf-epi and Born PSPI 

    allocate(kx2(pgrid%ax%n),ky2(pgrid%ay%n))
    allocate(tapx(pgrid%ntpx),tapy(pgrid%ntpy))

    select case (pgrid%mode)

      case ("imgfwd")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(bq(pgrid%ax%n,pgrid%ay%n),bqrf(pgrid%nqrf))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           if (igrid%cpximg) then
              allocate(cimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           else
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           end if

      case ("imgadj")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(bq(pgrid%ax%n,pgrid%ay%n),bqrf(pgrid%nqrf))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           if (igrid%cpximg) then
              allocate(cimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           else
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           end if

      case ("tomfwd")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
 !          allocate(dvel(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bq(pgrid%ax%n,pgrid%ay%n),bqrf(pgrid%nqrf))
           allocate(dq(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(dwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(wtmp_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(plan_2dtF(nth),plan_2dtI(nth))
           allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))

      case ("tomadj")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
 !          allocate(dvel(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bq(pgrid%ax%n,pgrid%ay%n),bqrf(pgrid%nqrf))
           allocate(dq(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(dwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(wtmp_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(plan_2dtF(nth),plan_2dtI(nth))
           allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))

      case ("tominv")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
   !        allocate(dvel(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bq(pgrid%ax%n,pgrid%ay%n),bqrf(pgrid%nqrf))
           allocate(dq(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(dwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(wtmp_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(plan_2dtF(nth),plan_2dtI(nth))
           allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           allocate(imag(igrid%ahx%n,igrid%ahy%n,igrid%ax%n,igrid%ay%n))

           if (hgrid%wnthes) then
              if (hgrid%hestyp==2) allocate(sfnc(wblock%ns))
           end if

           if (igrid%imgmsk) allocate(imsk(igrid%ax%n,igrid%ay%n))

      case ("hesian")
           allocate(sfnc(wblock%ns))
           allocate(csou(dgrid%ntrs,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(bq(pgrid%ax%n,pgrid%ay%n),bqrf(pgrid%nvrf))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))

           if (hgrid%hestyp==2) then
              allocate(crec(dgrid%ntrg,wblock%ns))
              allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
              allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
              allocate(plan_2drF(nth),plan_2drI(nth))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n)) 
           end if

    end select
    return
  end subroutine

  subroutine bwi_arrays_rset(ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak)
    integer :: ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak
    if (ogrid%ax%n/=ogrid_nx_bak .or. ogrid%ay%n/=ogrid_ny_bak .or. ogrid%az%n/=ogrid_nz_bak) then
       select case (pgrid%mode)
         case ("imgfwd")
              if (igrid%cpximg) then
                 deallocate(cimg)
                 allocate(cimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              else
                 deallocate(rimg)
                 allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              end if
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("imgadj")
              if (igrid%cpximg) then
                 deallocate(cimg)
                 allocate(cimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              else
                 deallocate(rimg)
                 allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              end if
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("tomfwd")
              deallocate(rimg)
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("tomadj")
              deallocate(rimg)
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("tominv")
              deallocate(rimg)
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("hesian")
              deallocate(rimg)
              deallocate(index_mxs,index_mys)
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              if (hgrid%hestyp==2) then
                 deallocate(index_mxr,index_myr)
                  allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
              end if
       end select
    end if
    return
  end subroutine

  subroutine bwi_module_init()
    integer :: stat

    call k2_map(kx2, pgrid%ax%n, pgrid%ax%d) 
    call k2_map(ky2, pgrid%ay%n, pgrid%ay%d)

    stat = bwi_tapers_init(tapx, tapy, pgrid%ntpx, pgrid%ntpy)

    select case (pgrid%mode)

      case ("imgfwd")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)

      case ("imgadj")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)

      case ("tomfwd")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2dtF, plan_2dtI, wtmp_slice, pgrid%ax%n, pgrid%ay%n, nth)

      case ("tomadj")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2dtF, plan_2dtI, wtmp_slice, pgrid%ax%n, pgrid%ay%n, nth)

      case ("tominv")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2dtF, plan_2dtI, wtmp_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fio_dq_init(qgrid, ftag_%grdo_)
           call bwi_fio_dq_init(qgrid, ftag_%grdp_)
           call bwi_fio_dq_init(qgrid, ftag_%srch_)

      case ("hesian")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           if (hgrid%hestyp==2) then
              call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)
           end if

    end select
    return
  end subroutine

end module
