module off2ang_kernfd_mod
  use  base_struct_mod
  use  base_cstret_mod
  use  base_fftwv3_mod
  use  base_fftsft_mod
  use  base_window_mod
  use  off2ang_tapers_mod
  implicit none

  type(off2ang)                         , private :: po2a
  type(axis)                            , private :: ahx, ahy, az
  type(axis)                            , private :: aagl_rfl, aagl_azm
  type(axis)                            , private :: akhx, akhy, akz
  complex, dimension(:,:)  , allocatable, private :: rttsfth 
  complex, dimension(:)    , allocatable, private :: rttsftv
  complex, dimension(:)    , allocatable, private :: c1dx, c1dz
  complex, dimension(:,:,:), allocatable, private :: c3d1, c3d2
  real   , dimension(:)    , allocatable, private :: kz, khx, agl_rfl, str
  real   , dimension(:)    , allocatable, private :: tpmz, tphx, tphy
  integer*8                             , private :: plan_1dxI
  integer*8                             , private :: plan_3dxF
  real   , parameter                    , private :: PI = 3.14159265359

  interface off2ang_kernfd_oper
    module procedure off2ang_kerfdc_oper
    module procedure off2ang_kerfdr_oper
  end interface

contains

  integer function off2ang_kernfd_init(ahx_in,ahy_in,aagl_rfl_in,aagl_azm_in,az_in,po2a_in) result(stat)
    type(axis)     :: ahx_in, ahy_in, aagl_rfl_in, aagl_azm_in, az_in 
    type(off2ang)  :: po2a_in
    integer        :: ntphx, ntphy, ntpmz

    real           :: shift
    integer        :: ikhx, iagl_rfl, ikz
    
    ahx      = ahx_in
    ahy      = ahy_in
    aagl_rfl = aagl_rfl_in
    aagl_azm = aagl_azm_in
    az       = az_in
    po2a     = po2a_in

    akhx%n = ahx%n+po2a%nhxpad; akhx%d = 2*PI/(akhx%n*ahx%d); akhx%o = (int(akhx%n/2)+2-akhx%n-1)*akhx%d
    akhy%n = ahy%n+po2a%nhypad; akhy%d = 2*PI/(akhy%n*ahy%d)
    akz%n  = az%n +po2a%nmzpad

    allocate(rttsfth(akhx%n,akhy%n))
    allocate(rttsftv(akz%n))
    allocate(c1dx(akhx%n))
    allocate(c1dz(akz%n))
    allocate(c3d1(akhx%n,akhy%n,akz%n))
    allocate(c3d2(akhx%n,akhy%n,akz%n))
    allocate(kz(akz%n))
    allocate(khx(akhx%n))
    allocate(agl_rfl(aagl_rfl%n))
    allocate(tphx(po2a%ntphx))
    allocate(tphy(po2a%ntphy))
    allocate(tpmz(po2a%ntpmz))

    if (po2a%push) then
       allocate(str(akhx%n))
    else
       allocate(str(aagl_rfl%n))
    end if

!   initialize FFTW
!   ---------------    
    call sfftw_dft_3dx_init(plan_3dxF, c3d1, akhx%n, akhy%n, akz%n, +1)
    call sfftw_dft_1dx_init(plan_1dxI, c1dz, akz%n, -1)

!   compute the kz index
!   --------------------
    call k_map(kz,akz%n,az%d)

!   compute the khx index
!   ---------------------
    do ikhx=1, akhx%n
       khx(ikhx) = akhx%o+(ikhx-1)*akhx%d
    end do

!   compute the angle index
!   -----------------------
    do iagl_rfl=1, aagl_rfl%n
       agl_rfl(iagl_rfl) = aagl_rfl%o+(iagl_rfl-1)*aagl_rfl%d
    end do

!   compute the horizontal shift
!   ----------------------------
    do ikhx=1, akhx%n
       shift           = -ahx%o*khx(ikhx)
       rttsfth(ikhx,1) = cmplx(cos(shift),sin(shift))
    end do

!   compute the vertical shift
!   --------------------------
    do ikz=1, akz%n
       shift        = -az%o*kz(ikz)
       rttsftv(ikz) = cmplx(cos(shift),sin(shift))
    end do

!   initialize the stretch
!   ----------------------
    stat = base_cstret_init(aagl_rfl, akhx, po2a%eps, po2a%push)

!   initialize the taper
!   --------------------
    stat = off2ang_tapers_init(tphx,tphy,tpmz,ahx%n,ahy%n,az%n,po2a%ntphx,po2a%ntphy,po2a%ntpmz)

    stat = 0
  end function

  integer function off2ang_kernfd_clos() result(stat)
    stat = 1
    deallocate(rttsfth)
    deallocate(rttsftv)
    deallocate(c1dx)
    deallocate(c1dz)
    deallocate(c3d1)
    deallocate(c3d2)
    deallocate(kz)
    deallocate(khx)
    deallocate(agl_rfl)
    deallocate(tphx)
    deallocate(tphy)
    deallocate(tpmz)

    if (po2a%push) then
       deallocate(str)
    else
       deallocate(str)
    end if
    stat = 0
  end function

  integer function off2ang_kerfdc_oper(gtha, gtho) result(stat)
    complex, dimension(:,:,:) :: gtha
    complex, dimension(:,:,:) :: gtho
    if (ahx%n>1) then
       call off2ang_kerfdc_2d(gtha, gtho)
    end if 
    stat = 0
  end function

  integer function off2ang_kerfdr_oper(gtha, gtho) result(stat)
    real, dimension(:,:,:) :: gtha
    real, dimension(:,:,:) :: gtho
    if (ahx%n>1) then
       call off2ang_kerfdr_2d(gtha, gtho)
    end if
    stat = 0
  end function

  subroutine off2ang_kerfdc_2d(gtha, gtho)
    complex, dimension(:,:,:) :: gtha
    complex, dimension(:,:,:) :: gtho

    integer                   :: ikz, iagl_rfl, ikhx, stat

!   empty the buffer
!   ----------------
    c3d1 = 0.; c3d2 = 0.

!   copy the offset gather
!   ----------------------
    c3d1(1:ahx%n,1:ahy%n,1:az%n) = gtho

!   apply tapers
!   ------------
    stat = off2ang_tapers_oper(c3d1(1:ahx%n,1:ahy%n,1:az%n))

!   3-D FFT
!   -------
    call sfftw_dft_3dx_oper(plan_3dxF, c3d1, akhx%n, akhy%n, akz%n)

!   flip khx and khy
!   ----------------
    do ikz=1, akz%n
       call base_fftsft_2d(c3d1(:,:,ikz), akhx%n, akhy%n)
    end do

!   apply phase shift
!   -----------------
    do ikz=1, akz%n
       c3d1(:,1,ikz) = c3d1(:,1,ikz)*rttsfth(:,1)*rttsftv(ikz)
    end do

!   apply the mapping
!   -----------------
    if (po2a%invt) then   !! inverting the nearest neighbor interpolation

       call off2ang_intinv2d(c3d1,c3d2,c1dx,khx,kz,str,akhx,aagl_rfl,akz,po2a%push)

    else  !! linear interpolation

       call off2ang_intadj2d(c3d1,c3d2,khx,kz,akhx,aagl_rfl,akz,po2a%push,po2a%intp)

    end if

!   apply the vertical shift back and 1-D IFFT
!   ------------------------------------------
    do iagl_rfl=1, aagl_rfl%n
       c1dz = c3d2(iagl_rfl,1,:)*conjg(rttsftv(:))
       call sfftw_dft_1dx_oper(plan_1dxI, c1dz, az%n)
       gtha(iagl_rfl,1,1:az%n) = c1dz(1:az%n)
    end do

    return
  end subroutine


  subroutine off2ang_kerfdr_2d(gtha, gtho)
    real   , dimension(:,:,:) :: gtha
    real   , dimension(:,:,:) :: gtho

    integer                   :: ikz, iagl_rfl, ikhx, stat

!   empty the buffer
!   ----------------
    c3d1 = 0.; c3d2 = 0.

!   copy the offset gather
!   ----------------------
    c3d1(1:ahx%n,1:ahy%n,1:az%n) = cmplx(gtho, 0.)

!   apply tapers
!   ------------
    stat = off2ang_tapers_oper(c3d1(1:ahx%n,1:ahy%n,1:az%n))

!   3-D FFT
!   -------
    call sfftw_dft_3dx_oper(plan_3dxF, c3d1, akhx%n, akhy%n, akz%n)

!   flip khx and khy
!   ----------------
    do ikz=1, akz%n
       call base_fftsft_2d(c3d1(:,:,ikz), akhx%n, akhy%n)
    end do

!   apply phase shift 
!   -----------------
    do ikz=1, akz%n
       c3d1(:,1,ikz) = c3d1(:,1,ikz)*rttsfth(:,1)*rttsftv(ikz)
    end do

!   apply the mapping
!   -----------------
    if (po2a%invt) then   !! inverting the nearest neighbor interpolation

       call off2ang_intinv2d(c3d1,c3d2,c1dx,khx,kz,str,akhx,aagl_rfl,akz,po2a%push)

    else  !! linear interpolation

       call off2ang_intadj2d(c3d1,c3d2,khx,kz,akhx,aagl_rfl,akz,po2a%push,po2a%intp)

    end if

!   apply vertical shift back and 1-D IFFT
!   --------------------------------------
    do iagl_rfl=1, aagl_rfl%n
       c1dz = c3d2(iagl_rfl,1,:)**conjg(rttsftv(:))
       call sfftw_dft_1dx_oper(plan_1dxI, c1dz, az%n)
       gtha(iagl_rfl,1,1:az%n) = real(c1dz(1:az%n))
    end do
    return
  end subroutine

  subroutine off2ang_intinv2d(c3d1,c3d2,c1dx,khx,kz,str,akhx,aagl_rfl,akz,push)
    implicit none
    complex, dimension(:,:,:) :: c3d1, c3d2
    complex, dimension(:)     :: c1dx
    real   , dimension(:)     :: khx, kz, str
    type(axis)                :: akhx, aagl_rfl, akz
    logical                   :: push
 
    integer                   :: ikz, ikhx, iagl_rfl, stat

    c3d2 = 0.

    if (push) then !! push
       do ikz=1, akz%n
          str = 0.
          if (kz(ikz)/=0.) then
             do ikhx=1, akhx%n
                str(ikhx)  = - atan(khx(ikhx)/kz(ikz))*180./PI
             end do
          else
             str = 90.
          end if
          c1dx = 0.
          stat  = base_cstret_oper(str, c1dx, c3d1(:,1,ikz))
          c3d2(:,1,ikz) = c1dx
       end do

    else !! pull
       do ikz=1, akz%n
          str = 0.
          do iagl_rfl=1, aagl_rfl%n
             str(iagl_rfl) = - kz(ikz)*tan(agl_rfl(iagl_rfl)*PI/180.)
          end do
          c1dx = 0.
          stat  = base_cstret_oper(str, c1dx, c3d1(:,1,ikz))
          c3d2(:,1,ikz) = c1dx
       end do

    end if
    return
  end subroutine

  subroutine off2ang_intadj2d(c3d1,c3d2,khx,kz,akhx,aagl_rfl,akz,push,intp)
    complex, dimension(:,:,:) :: c3d1, c3d2
    real   , dimension(:)     :: khx, kz 
    type(axis)                :: akhx, aagl_rfl, akz
    logical                   :: push
    integer                   :: intp

    integer                   :: ikz, iagl_rfl, ikhx
    integer                   :: jkhxm, jkhxp
    integer                   :: jagl_rflm, jagl_rflp
    real                      :: khx_new, agl_rfl_new 
    real                      :: rkhx, fkhxp, fkhxm
    real                      :: ragl_rfl, fagl_rflp, fagl_rflm


    c3d2 = 0.

    if (push) then
       do ikz=1, akz%n
          if (kz(ikz)==0.) cycle
          do ikhx=1, akhx%n
             agl_rfl_new = - atan(khx(ikhx)/kz(ikz))*180/PI

             select case (intp) !! nearest neighbor inverpolation
               case (1)
                    iagl_rfl = 1.5 + (agl_rfl_new-aagl_rfl%o)/aagl_rfl%d
                    if (iagl_rfl<1 .or. iagl_rfl>aagl_rfl%n) cycle
                    c3d2(iagl_rfl,1,ikz) = c3d2(iagl_rfl,1,ikz) + c3d1(ikhx,1,ikz)

               case (2)              !! linear interpolation
                    ragl_rfl  = (agl_rfl_new-aagl_rfl%o)/aagl_rfl%d
                    jagl_rflm = ragl_rfl
                    if (jagl_rflm<1 .or. jagl_rflm>aagl_rfl%n-1) cycle
                    fagl_rflp = ragl_rfl - jagl_rflm
                    fagl_rflm = 1.-fagl_rflp
                    c3d2(jagl_rflm,1,ikz) = c3d2(jagl_rflm,1,ikz)+c3d1(ikhx,1,ikz)*fagl_rflm
                    c3d2(jagl_rflp,1,ikz) = c3d2(jagl_rflp,1,ikz)+c3d1(ikhx,1,ikz)*fagl_rflp
             end select

          end do
       end do

    else
       do ikz=1, akz%n
          do iagl_rfl=1, aagl_rfl%n
             khx_new = -kz(ikz)*tan(agl_rfl(iagl_rfl)*PI/180.)

             select case (intp) !! nearest neighbor interpolation 
               case (1)
                    ikhx = 1.5+(khx_new-akhx%o)/akhx%d
                    if (ikhx<1 .or. ikhx>akhx%n) cycle
                    c3d2(iagl_rfl,1,ikz) = c3d2(iagl_rfl,1,ikz)+c3d1(ikhx,1,ikz) 

               case (2)         !! linear interpolation
                    rkhx  = (khx_new-akhx%o)/akhx%d
                    jkhxm = rkhx
                    if (jkhxm<1 .or. jkhxm>akhx%n-1) cycle
                    fkhxp = rkhx-jkhxm
                    fkhxm = 1.-fkhxp
                    c3d2(iagl_rfl,1,ikz) = c3d2(iagl_rfl,1,ikz)+c3d1(jkhxm,1,ikz)*fkhxm+c3d1(jkhxp,1,ikz)*fkhxp
             end select

          end do
       end do  
    end if
    return 
  end subroutine

  subroutine k_map(k,n,d)
    real,  dimension(:) :: k
    integer, intent(in) :: n
    real,    intent(in) :: d
    integer             :: i,m

    k=0.
    if(n/=1) then
       m = n/2
       do i=1,m+1
          k(i)=i-1
       end do
       do i=m+2,n
          k(i)=i-n-1
       end do
    else
       k=0. !! make sure k=0 if n=1
    end if

    k = k / (n*d)      !! 1/length
    k = 2 * PI * k     !! radians
    return
  end subroutine
end module
